/*
 * File: LCD_graphics_lib.h
 *
 * API Library definitions for Monochrome Graphic LCD module, 128 x 64 pixels.
 */
 
#ifndef LCD_GRAPHICS_LIB_H
#define LCD_GRAPHICS_LIB_H

#ifdef USE_LCD_CONTROLLER_ST7920 
#include "Drivers/LCD_ST7920_drv.h"
#else  // assume USE_LCD_CONTROLLER_KS0108
#include "Drivers/LCD_KS0108_drv.h"
#endif

// Rendering modes for LCD write functions...
#define CLEAR_PIXELS          0
#define SET_PIXELS            1
#define FLIP_PIXELS           2

//#define SUPPORT_12PT_AND_24PT_FONTS    // Maybe someday! - MJB

// Character font styles;  size is cell height in pixels
// Use one of the font names defined here as the arg value in function LCD_SetFont(arg)
//
enum  Graphics_character_fonts
{
    BAUER_MONO_8_NORM = 0,   // Mono-spaced font;  cell width is 5 pix
    BAUER_MONO_8_BOLD,       // N/A - same as MONO_8_NORM (not bold)
    BAUER_PROP_8_NORM,       // Proportional font;  cell width is 3..5 pix
    BAUER_PROP_8_BOLD,       // N/A - same as PROP_8_NORM (not bold)

    BAUER_MONO_12_NORM,      // N/A -- use BAUER_PROP_12_NORM
    BAUER_MONO_12_BOLD,      // N/A -- use BAUER_PROP_12_BOLD
    BAUER_PROP_12_NORM,      // Proportional font;  cell width is 4..7 pix
    BAUER_PROP_12_BOLD,      // as above, but bold weight

    BAUER_MONO_16_NORM,      // Mono-spaced font;  cell width is 10 pix
    BAUER_MONO_16_BOLD,      // as above, but bold weight
    BAUER_PROP_16_NORM,      // Proportional font;  cell width is 6..10 pix
    BAUER_PROP_16_BOLD,      // as above, but bold weight

    BAUER_MONO_24_NORM,      // N/A -- use BAUER_PROP_24_NORM
    BAUER_MONO_24_BOLD,      // N/A -- use BAUER_PROP_24_BOLD
    BAUER_PROP_24_NORM,      // Proportional font;  cell width is 8..14 pix
    BAUER_PROP_24_BOLD       // as above, but bold weight
};


typedef  const unsigned char  bitmap_t;  // Bitmap image is an array of bytes in flash

//---------- LCD controller low-level functions, defined in driver module ---------------
//
extern  void    LCD_ClearGDRAM();
extern  void    LCD_WriteBlock(uint16 *scnBuf, uint16 x, uint16 y, uint16 w, uint16 h);

//---------- LCD Function & macro library (API) -----------------------------------------
//
#define LCD_GetMaxX()       (127)
#define LCD_GetMaxY()       (63)

bool   LCD_Init(void);                   // LCD controller initialisation
void   LCD_ClearScreen(void);            // Clear LCD GDRAM and MCU RAM buffers
void   LCD_Mode(uint8 mode);             // Set pixel write mode (set, clear, flip)

void     LCD_PosXY(uint16 x, uint16 y);  // Set graphics cursor position to (x, y)
uint16   LCD_GetX(void);                 // Get graphics cursor pos x-coord
uint16   LCD_GetY(void);                 // Get graphics cursor pos y-coord

void     LCD_BlockFill(uint16 w, uint16 h);  // Fill area, w x h pixels, at cursor (x, y)
uint8    LCD_PutImage(bitmap_t *image, uint16 w, uint16 h);   // Show bitmap image at (x, y)
uint16  *LCD_ScreenCapture();            // Return a pointer to the screen buffer

void   LCD_SetFont(uint8 font_ID);       // Set font for char or text display
uint8  LCD_GetFont();                    // Get current font ID
void   LCD_PutChar(uint8 uc);            // Show ASCII char at (x, y)
void   LCD_PutText(uint8 *str);          // Show text string at (x, y)
void   LCD_PutDigit(uint8 bDat);         // Show hex/decimal digit value (1 char)
void   LCD_PutHexByte(uint8 bDat);       // Show hexadecimal byte value (2 chars)
void   LCD_PutDecimalWord(uint16 val, uint8 fieldSize);

// These macros draw various objects at the current graphics cursor position...
#define LCD_PutPixel()           LCD_BlockFill(1, 1)
#define LCD_DrawBar(w, h)        LCD_BlockFill(w, h)
#define LCD_DrawLineHoriz(len)   LCD_BlockFill(len, 1)
#define LCD_DrawLineVert(len)    LCD_BlockFill(1, len)

//---------------------------------------------------------------------------------------

void  LCD_TestCharGen(uint8 fontID);     // Test character generator functions
void  LCD_TestWriteBlock();              // Test low-level driver functions

#endif  // LCD_GRAPHICS_LIB_H
