/**************************************************************************************************
 *
 * FileName:      data_flash_fs.c
 *
 * Module implements a primitive file system for serial data-flash (SDF) memory devices,
 * providing a more convenient suite of API functions than the low-level SST25VF016 module.
 * The application should not need to call any functions in module "SST25VF016.c" directly.
 *
 * This version is customised to suit the SST25VFxxx SPI flash device family.
 *
 * Dependencies:  "../Drivers/SST25VF016.c", "../Drivers/SPI_drv.c"
 *
 * Originated:    April 2014, M.J.Bauer  [www.mjbauer.biz]
 *
 * ================================================================================================
 *
 * The SDF file system allows files of various sizes to be written to and retrieved from
 * a serial data-flash memory device interfaced to the MCU via an SPI channel.
 * The system is designed for "write once, read-only memory" (WORM) applications, e.g.
 * for storage of PCM sound clip data, persistent parameters, etc.
 * Once a file is written, it cannot be removed without erasing the entire device; except
 * that the memory sector(s) allocated to a file may be erased and re-written without
 * changing the file's directory entry (FIR) -- [see SDF_ferase(), SDF_reopen_WR()].
 * A file can also be "hidden" if it is not wanted -- [see SDF_fhide()].
 *
 * One or more sector(s) at the start of the SDF memory is/are reserved for the 'File
 * Directory Block' (FDB) which is an array of 'File Index Record' (FIR) structures each
 * of 32 bytes containing information about a stored file, i.e. its location in the "volume",
 * the file size, file name, etc, and a link to the first sector of the next file.
 * Only one file can be open at any time for byte-sequential writing or reading. However,
 * there is no need to "open" files for random access with SDF_ReadArray(), SDF_WriteArray().
 *
 * ================================================================================================
 */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "data_flash_fs.h"

#if USE_CONSOLE_CLI
#include "console.h"

static void   dump_page_buffer(uint8 *pbAddr, int iFlashAddr);
static uint8  fill_flash_4K_block(uint32 lAddr);
static int    verify_flash_4K_block( uint32 lAddr );

#endif

// File Index Record (FIR) of currently open file (local RAM copy)
static DataFlash_FileIndexRecord_t  m_SST25_FileInfo;

static uint8  isFileOpenRD;   // File is open for byte-sequential reading
static uint8  isFileOpenWR;   // File is open for byte-sequential writing
static uint8  isFileErasing;  // File erase operation started by SDF_ferase() is in train

static uint32 byteAddress;    // for byte-sequential read or write of open file
static int32  byteCount;      //  ..   ..    ..    ..    ..


/**************************************************************************************************
 * Function:     SDF_format()
 *
 * Overview:     Initialize serial data-flash (SDF) memory device for first-time use.
 *               The entire device is erased.
 *
 * Return val:   TRUE if the operation succeeded, else FALSE (0).
 */
uint8  SDF_format(void)
{
    SST25ResetWriteProtection();
    SST25ChipErase();

    isFileOpenRD = 0;
    isFileOpenWR = 0;
    isFileErasing = 0;

    return 1;
}


/**************************************************************************************************
 * Function:      SDF_fopen_WR()
 *
 * Overview:      Open SDF file for writing in byte-sequential mode.
 *
 * Entry arg(s):  fileName = pointer to file-name string (1 char min., 19 chars max.)
 *                fileSize = maximum (pref. exact) number of data bytes to be written
 *
 * Return val:    file index number if the operation succeeded, else ERROR (-1).
 */
int  SDF_fopen_WR(char *fileName, int32 fileSize)
{
    int   findex;
    int   sectorsNeeded;
    int   sectorsFree;
    int   nextFreeSector;
    uint32  address = 0;

    if (isFileOpenWR || isFileOpenRD || isFileErasing) return ERROR;

    if (fileName[0] <= 0x20) return ERROR;  // Illegal filename

    sectorsNeeded = fileSize / SST25_SECTOR_SIZE;
    if (fileSize % SST25_SECTOR_SIZE) sectorsNeeded++;

    // Find the next free (blank/erased) file index record in the directory...
    for (findex = 0; findex < SDF_MEMORY_MAX_FILES; findex++)
    {
        address = (uint32) findex * sizeof(m_SST25_FileInfo);
        SST25ReadArray(address, (uint8 *) &m_SST25_FileInfo, sizeof(m_SST25_FileInfo));

        if (m_SST25_FileInfo.fileIndex == 0xFFFF) break;  // Free entry found
    }

    if (findex < SDF_MEMORY_MAX_FILES)   // If enough room left in directory...
    {
        if (findex == 0)  // Directory empty, no files written yet...
        {
            nextFreeSector = 1;   // assuming directory uses 1 sector
            sectorsFree = SST25_TOTAL_SECTORS - 1;
        }
        else  // Read file info from last FIR written in the directory...
        {
            address = (uint32) (findex - 1) * sizeof(m_SST25_FileInfo);
            SST25ReadArray(address, (uint8 *) &m_SST25_FileInfo, sizeof(m_SST25_FileInfo));
            nextFreeSector = m_SST25_FileInfo.linkSector;
            sectorsFree = SST25_TOTAL_SECTORS - nextFreeSector - 1;
        }

        if (sectorsNeeded < sectorsFree)   // If enough room left in volume...
        {
            // Find starting address for first call to SDF_fwrite()
            byteAddress = (uint32) nextFreeSector * SST25_SECTOR_SIZE;
            byteCount = 0;

            // Populate directory entry (FIR) with file info of newly opened file
            m_SST25_FileInfo.fileIndex = findex;
            m_SST25_FileInfo.begSector = nextFreeSector;
            m_SST25_FileInfo.linkSector = nextFreeSector + sectorsNeeded;
            m_SST25_FileInfo.fileSize = fileSize;
            strncpy(m_SST25_FileInfo.fileName, fileName, 20);
            m_SST25_FileInfo.fileName[19] = 0;  // NUL term.

            // Write file info record of newly opened file into the directory
            address = (uint32) findex * sizeof(m_SST25_FileInfo);
            SST25WriteArray(address, (uint8 *) &m_SST25_FileInfo, sizeof(m_SST25_FileInfo));

            // Read back and verify that the file index was written
            SST25ReadArray(address, (uint8 *) &m_SST25_FileInfo, 4);
            if (m_SST25_FileInfo.fileIndex == findex) isFileOpenWR = 1; // success
            else findex = ERROR;
        }
        else findex = ERROR;
    }
    else findex = ERROR;

    return findex;
}


/**************************************************************************************************
 * Function:      SDF_fwrite()
 *
 * Overview:      Write (append) one byte to a file in serial data-flash (SDF) memory.
 *                The file must be open for writing by a prior call to SDF_fopen_WR()
 *                or SDF_freopen_WR().
 *                The function returns FALSE (0) when the number of bytes written reaches
 *                the file size specified in the directory (FIR).
 *
 * Entry arg(s):  bDat   = data byte to be written
 *
 * Return val:    TRUE if the operation succeeded, else FALSE (0).
 */
uint8  SDF_fwrite(uint8 bDat)
{
    uint8  status = 0;

    if (isFileOpenWR && byteCount < m_SST25_FileInfo.fileSize)
    {
        SST25WriteByte(bDat, byteAddress);
        byteAddress++;
        byteCount++;
        status = 1;
    }

    return status;
}


/**************************************************************************************************
 * Function:      SDF_find()
 *
 * Overview:      Find a file in the directory of serial data-flash (SDF) memory.
 *                The function returns the file index of the first file found in the directory
 *                matching the given file-name, or ERROR (-1) if no match is found.
 *                The file-name comparison is case-sensitive.
 *
 * Entry arg(s):  fileName = pointer to file-name string (19 chars max.)
 *
 * Return val:    file index number if the operation succeeded, else ERROR (-1).
 */
int  SDF_find(char *fileName)
{
    DataFlash_FileIndexRecord_t  *pFileInfo;
    int  file_ID = ERROR;
    int  i;

    for (i = 0;  i < SDF_MEMORY_MAX_FILES;  i++)
    {
        pFileInfo = SDF_finfo(i);
        if (pFileInfo == NULL)  break;  // end of directory

        if (strncmp(pFileInfo->fileName, fileName, 20) == 0)  // match
        {
            file_ID = pFileInfo->fileIndex;
            break;
        }
    }

    return file_ID;
}


/**************************************************************************************************
 * Function:      SDF_fopen_RD()
 *
 * Overview:      Open SDF file for reading in byte-sequential mode.
 *
 * Entry arg(s):  fileID = file index (0..SDF_MEMORY_MAX_FILES)
 *
 * Return val:    TRUE if the operation succeeded, else FALSE (0).
 */
uint8  SDF_fopen_RD(uint16 fileID)
{
    uint8   status = 0;
    uint32  address = (uint32) fileID * sizeof(m_SST25_FileInfo);

    if (!isFileOpenRD && !isFileOpenWR && !isFileErasing && fileID < SDF_MEMORY_MAX_FILES)
    {
        SST25ReadArray(address, (uint8 *) &m_SST25_FileInfo, sizeof(m_SST25_FileInfo));

        if (m_SST25_FileInfo.fileIndex == fileID)  // Valid FIR found (file exists)
        {
            byteAddress = (uint32) m_SST25_FileInfo.begSector * SST25_SECTOR_SIZE;
            byteCount = 0;
            isFileOpenRD = 1;
            status = 1;
        }
    }

    return status;
}


/**************************************************************************************************
 * Function:      SDF_fread()
 *
 * Overview:      Read next byte (sequentially) from file in serial data-flash (SDF) memory.
 *                A file must be open for reading by a prior call to SDF_fopen_RD().
 *
 * Return val:    Next unread byte from SDF file, or, if no data available, EOF (-1).
 */
int  SDF_fread(void)
{
    int  retValue = (-1);

    if (isFileOpenRD && byteCount < m_SST25_FileInfo.fileSize)
    {
        if (byteCount == 0)  // First call following SDF_fopen_RD()
            retValue = SST25ReadFirstByte(byteAddress);
        else
            retValue = SST25ReadNextByte(TRUE);
        byteCount++;
    }

    return retValue;
}


/**************************************************************************************************
 * Function:      SDF_fclose()
 *
 * Overview:      Close any file which is open for read or write in byte-sequential mode.
 *                Terminates any incomplete file erase operation started by SDF_ferase().
 *
 * Note:          If a file is open for writing, and the number of bytes written is less
 *                than the specified file size when the file is closed, then the file size
 *                will remain as specified and "unwritten" bytes will remain blank (0xFF).
 *
 * Return val:    TRUE if the operation succeeded, else FALSE (0).
 */
uint8  SDF_fclose(void)
{
    if (isFileOpenRD) SST25ReadNextByte(FALSE);  // Release SPI slave

    isFileOpenWR = 0;
    isFileOpenRD = 0;
    isFileErasing = 0;

    return TRUE;
}


/**************************************************************************************************
 * Function:      SDF_finfo()
 *
 * Overview:      Return file info from FIR in serial data-flash (SDF) memory directory.
 *                It does not matter if a file is open or not when this function is called.
 *
 * Entry arg(s):  fileID = file index (0..SDF_MEMORY_MAX_FILES)
 *
 * Return val:    Pointer to FIR structure at the specified file index, or...
 *                NULL if the file index is invalid, or some other error occurs.
 */
DataFlash_FileIndexRecord_t *SDF_finfo(uint16 fileID)
{
    static  DataFlash_FileIndexRecord_t  SDF_FileInfo;

    DataFlash_FileIndexRecord_t  *pFIR = &SDF_FileInfo;
    uint32  address = (uint32) fileID * sizeof(DataFlash_FileIndexRecord_t);

    SST25ReadArray(address, (uint8 *) &SDF_FileInfo, sizeof(DataFlash_FileIndexRecord_t));

    if (SDF_FileInfo.fileIndex != fileID) pFIR = NULL;  // error: past end of directory
    if (fileID >= SDF_MEMORY_MAX_FILES) pFIR = NULL;    // error: file index out of range

    return pFIR;
}


/**************************************************************************************************
 * Function:      SDF_fhide()
 *
 * Overview:      This operation prevents the file name from showing up in directory
 *                listings. This is a way to virtually remove an unwanted file, although
 *                its directory entry (FIR) and file data remain intact.
 *                The function simply sets the first char of the filename to NUL (0).
 *
 * Entry arg(s):  fileID = file index (0..SDF_MEMORY_MAX_FILES)
 *
 * Return val:    TRUE if the operation succeeded, else FALSE (0).
 */
uint8  SDF_fhide(uint16 fileID)
{
    DataFlash_FileIndexRecord_t  SDF_FileInfo;  // local copy of FIR
    uint8   status = 0;
    uint32  address = (uint32) fileID * sizeof(DataFlash_FileIndexRecord_t);

    if (fileID < SDF_MEMORY_MAX_FILES)
    {
        SST25ReadArray(address, (uint8 *) &SDF_FileInfo, sizeof(DataFlash_FileIndexRecord_t));

        if (SDF_FileInfo.fileIndex == fileID)  // Valid FIR found (file exists)
        {
            // Clear filename and write FIR back into the directory
            SDF_FileInfo.fileName[0] = 0;  // NUL term.
            SST25WriteArray(address, (uint8 *) &SDF_FileInfo, sizeof(DataFlash_FileIndexRecord_t));
            status = 1;
        }
    }

    return status;
}


/**************************************************************************************************
 * Function:      SDF_ferase()
 *
 * Overview:      Erase the contents of an existing file without changing its directory entry;
 *                i.e. the file name, file size and number of sectors allocated remain intact.
 *                To avoid blocking (delaying) other processes, this function must be called
 *                repetitively until all sectors in the file have been erased.
 *                It is advisable to impose a time-out or limit the number of calls made.
 *                The caller should allow an erase time of up to 25ms per sector.
 *
 * Entry arg(s):  fileID = file index (0..SDF_MEMORY_MAX_FILES)
 *
 * Return val:    (int) Number of sectors remaining to be erased (zero when done) ...OR...
 *                      ERROR (-1) if the file cannot be erased, invalid fileID, etc.
 */
int  SDF_ferase(uint16 fileID)
{
    static  int  sectorsRemaining;  // or ERROR
    static  uint32  address;
    DataFlash_FileIndexRecord_t  SDF_FileInfo;  // local copy of FIR

    if (!isFileErasing)  // First call -- begin erase
    {
        if (!isFileOpenRD && !isFileOpenWR && fileID < SDF_MEMORY_MAX_FILES)
        {
            address = (uint32) fileID * sizeof(DataFlash_FileIndexRecord_t);
            SST25ReadArray(address, (uint8 *) &SDF_FileInfo, sizeof(DataFlash_FileIndexRecord_t));

            if (SDF_FileInfo.fileIndex == fileID)  // Valid FIR found (file exists)
            {
                // Find flash address at beginning of first sector to erase...
                address = (uint32) SDF_FileInfo.begSector * SST25_SECTOR_SIZE;
                sectorsRemaining = SDF_FileInfo.linkSector - SDF_FileInfo.begSector;
                isFileErasing = 1;
            }
            else  sectorsRemaining = ERROR;
        }
        else  sectorsRemaining = ERROR;
    }

    if (isFileErasing)  // File erase in progress
    {
        SST25SectorErase(address);  // may take up to 25ms (typ. 18ms for SST25VF016)
        sectorsRemaining-- ;
        if (sectorsRemaining == 0) isFileErasing = 0;  // done!
        else  address += SST25_SECTOR_SIZE;  // start addr of next sector
    }

    return sectorsRemaining;  // or ERROR
}


/**************************************************************************************************
 * Function:      SDF_freopen_WR()
 *
 * Overview:      Re-open an existing file (assumed to be erased) for re-writing in
 *                byte-sequential mode.
 *                The file size remains unchanged, so that SDF_fread() will continue to
 *                return data, up to the fixed file size, regardless of how many bytes
 *                have been re-written.  [See also: SDF_ferase() and SDF_fclose().]
 *
 * Entry arg(s):  fileID = file index (up to SDF_MEMORY_MAX_FILES)
 *
 * Return val:    TRUE if the operation succeeded;  otherwise FALSE...
 *                (e.g. file not erased, file already open, invalid fileID, etc).
 */
uint8  SDF_freopen_WR(uint16 fileID)
{
    uint32  address = (uint32) fileID * sizeof(m_SST25_FileInfo);

    if (isFileErasing || isFileOpenWR || isFileOpenRD || fileID >= SDF_MEMORY_MAX_FILES)
    {
        return FALSE;
    }

    // Read file info from FIR in directory block...
    SST25ReadArray(address, (uint8 *) &m_SST25_FileInfo, sizeof(m_SST25_FileInfo));

    if (m_SST25_FileInfo.fileIndex == fileID)  // Valid FIR found (file exists)
    {
        // Find starting address for first call to SDF_fwrite()
        byteAddress = (uint32) m_SST25_FileInfo.begSector * SST25_SECTOR_SIZE;
        byteCount = 0;
        isFileOpenWR = 1;
        return TRUE;
    }
    else return FALSE;  // fileID not found in FIR
}


/**************************************************************************************************
 * Function:      SDF_WriteArray()
 *
 * Overview:      Write a specified number of bytes into an existing file in SDF memory
 *                from an array in MCU data space (RAM) using "random access".
 *                It is assumed that the target area to be written is already erased.
 *                The function cannot be called while any file is open for sequential read
 *                or write access.
 *
 *                If the available space in the target file is less than the required number
 *                of bytes (byteCount), only the data which can fit will be written.
 *                The function exits immediately if a byte written fails to verify.
 *
 * Entry arg(s):  fileID     = file index (up to SDF_MEMORY_MAX_FILES)
 *                pSourceBuf = pointer to source array in MCU RAM
 *                offset     = address of first byte to be written into SDF memory,
 *                               relative to start address of SDF file (@ fileID).
 *                byteCount  = number of bytes to be written
 *
 * Return val:    Number of bytes written to SDF memory and verified.
 */
long  SDF_WriteArray(uint16 fileID, uint8 *pSourceBuf, long offset, long byteCount)
{
    uint32  sdf_max_address;
    uint32  sdf_address;
    long    bytesWritten = 0;
    DataFlash_FileIndexRecord_t  SDF_FileInfo;  // local copy of FIR

    // Read file info from FIR in directory block...
    sdf_address = (uint32) fileID * sizeof(DataFlash_FileIndexRecord_t);
    SST25ReadArray(sdf_address, (uint8 *) &SDF_FileInfo, sizeof(DataFlash_FileIndexRecord_t));

    if (SDF_FileInfo.fileIndex == fileID)  // Valid FIR found (file exists)
    {
        // Find address for first byte to be written...
        sdf_address = (uint32) SDF_FileInfo.begSector * SST25_SECTOR_SIZE + offset;
        sdf_max_address = (uint32) SDF_FileInfo.linkSector * SST25_SECTOR_SIZE - 1;

        while (bytesWritten < byteCount)
        {
            if (sdf_address > sdf_max_address) break;  // Not enough room!
            ////
            SST25WriteByte(*pSourceBuf, sdf_address);
            if (SST25ReadByte(sdf_address) != *pSourceBuf) break;  // Verify failed!
            ////
            pSourceBuf++;  sdf_address++;  bytesWritten++;
        }
    }

    return bytesWritten;
}


/**************************************************************************************************
 * Function:      SDF_ReadArray()
 *
 * Overview:      Read a specified number of bytes from an existing file in SDF memory
 *                into an array in MCU data space using "random access".
 *                The function cannot be called while any file is open for sequential read
 *                or write access.
 *
 *                If 'end-of-file' is encountered before the required number of bytes has
 *                been copied, the function exits with return value less than 'byteCount'.
 *
 * Entry arg(s):  fileID     = file index (up to SDF_MEMORY_MAX_FILES)
 *                pDestinBuf = pointer to destination array in MCU RAM
 *                offset     = address of first byte to be read from SDF memory,
 *                               relative to start address of SDF file (@ fileID).
 *                byteCount  = number of bytes to be read
 *
 * Return val:    Number of bytes copied from SDF memory into the destination array.
 */
long  SDF_ReadArray(uint16 fileID, uint8 *pDestinBuf, long offset, long byteCount)
{
    uint32  sdf_max_address;
    uint32  sdf_address;
    long    bytesCopied = 0;
    DataFlash_FileIndexRecord_t  SDF_FileInfo;  // local copy of FIR

    // Read file info from FIR in directory block...
    sdf_address = (uint32) fileID * sizeof(DataFlash_FileIndexRecord_t);
    SST25ReadArray(sdf_address, (uint8 *) &SDF_FileInfo, sizeof(DataFlash_FileIndexRecord_t));

    if (SDF_FileInfo.fileIndex == fileID)  // Valid FIR found (file exists)
    {
        // Find address for first byte to be written...
        sdf_address = (uint32) SDF_FileInfo.begSector * SST25_SECTOR_SIZE + offset;
        sdf_max_address = (uint32) SDF_FileInfo.linkSector * SST25_SECTOR_SIZE - 1;

        while (bytesCopied < byteCount)
        {
            if (sdf_address > sdf_max_address) break;  // End-of-file!
            ////
            if (bytesCopied == 0) *pDestinBuf = SST25ReadFirstByte(sdf_address);
            else  *pDestinBuf = SST25ReadNextByte(TRUE);  // Auto address inc.
            ////
            pDestinBuf++;  sdf_address++;  bytesCopied++;
        }

        if (bytesCopied) SST25ReadNextByte(FALSE);  // Release the SPI port
    }

    return bytesCopied;
}



#if USE_CONSOLE_CLI
/*^***********************************************************************************************
 *
 * CLI command function:  Cmnd_sdf
 *
 * This command is for managing files on the serial data-flash (SDF) memory device.
 */
void  Cmnd_sdf(int argCount, char * argVal[])
{
    DataFlash_FileIndexRecord_t  *pFileInfo;
    short   file_ID;
    uint16  lastLinkSector;
    int     i, fileCount = 0;
    int     hiddenCount = 0;
    int     sizeSectors;
    int     sectorsUsed = 0;
    char    cOption;
    char    foundError = 0;

    if ( argCount == 1 || argVal[1][0] == '?')
    {
        putstr( "Usage:  sdf  <opt> [arg2] [arg3] \n" );
        putstr( "<opt> = \n" );
        putstr( " -f : Format (erase) SDF memory, arg2 = '+++' \n");
        putstr( " -l : List SDF directory (file index) \n");
        putstr( " -h : Hide file, arg2 = SDF fid# \n");
        putstr( " -m : Make file, arg2 = name, arg3 = size (sectors)\n");
        putstr( " -k : ChecK SDF file system integrity \n");
        return;
    }

    cOption = tolower( argVal[1][1] );

    if (argCount >= 3) file_ID = atoi(argVal[2]);

    switch ( cOption )
    {
    case 'f':  // Format (erase) SDF memory
    {
        if (!SuperUserAccess()) break;

        if (argCount == 3 && strmatch(argVal[2], "+++")) SDF_format();
        else  putstr("! Wrong syntax. (Need arg2 = '+++') \n");
        break;
    }
    case 'k':  // Check SDF file system (directory)
    {
        for (i = 0;  i < SDF_MEMORY_MAX_FILES;  i++)
        {
            pFileInfo = SDF_finfo(i);
            if (pFileInfo == NULL)  break;  // end of directory, or bad FIR

            if ( pFileInfo->fileName[0] == 0xFF
            ||   (pFileInfo->begSector & 0xFFF) == 0xFFF
            ||   (pFileInfo->linkSector & 0xFFF) == 0xFFF
            ||   pFileInfo->linkSector <= pFileInfo->begSector )
            {
                putstr("Bad FIR in file #: ");
                putDecimal(i, 1);  putstr("\n");
                foundError = 1;
                break;
            }
            lastLinkSector = pFileInfo->linkSector;
            fileCount++;
        }

        putstr("Directory check ");
        if (!foundError) putstr("passed...\n");  else putstr("failed...\n");
        putstr("Intact FIR's found:  ");  putDecimal(fileCount, 1);  putstr("\n");
        putstr("Last link sector #: $");  putHexWord(lastLinkSector);  putstr("\n");
        break;
    }
    case 'l':  // List file directory
    {
        putstr(" Fid  $BegAddr  #Sect   #Bytes  FileName  \n");

        for (i = 0;  i < SDF_MEMORY_MAX_FILES;  i++)
        {
            pFileInfo = SDF_finfo(i);
            if (pFileInfo == NULL)  break;  // end of directory, or bad FIR

            fileCount++;
            sizeSectors = pFileInfo->linkSector - pFileInfo->begSector;
            sectorsUsed += sizeSectors;

            if (pFileInfo->fileName[0] > 0x20)   // file is not hidden
            {
                putDecimal(pFileInfo->fileIndex, 4);
                putstr("   ");
                putHexWord(pFileInfo->begSector);
                putstr("000  ");
                putDecimal(sizeSectors, 4);
                putstr("   ");
                putDecimal(pFileInfo->fileSize, 7);  // bytes
                putstr("  ");  putstr(pFileInfo->fileName);
                putstr("\n");
            }
            else  hiddenCount++;
        }
        // Show SDF usage stats...
        putDecimal(fileCount, 1);  putstr(" files found (");
        putDecimal(hiddenCount, 1);  putstr(" hidden).  ");
        putDecimal(sectorsUsed, 1);  putstr("/");
        putDecimal(SST25_TOTAL_SECTORS - sectorsUsed, 1);
        putstr(" sectors used/free.\n");
        break;
    }
    case 'h':  // Hide file -- clear file name
    {
        if (argCount == 3)
        {
            file_ID = atoi(argVal[2]);
            if (SDF_fhide(file_ID) == FALSE)  putstr("! Failed... Check file ID.\n");
        }
        else putstr( "! Need file_ID number.\n" );
        break;
    }
    case 'm':  // Make new (empty) file
    {
        char  fileName[24];
        int32 fileSize;

        strncpy(fileName, argVal[2], 19);
        fileName[19] = 0;
        fileSize = atoi(argVal[3]) * SST25_SECTOR_SIZE;  // bytes
        file_ID = ERROR;

        if (fileSize > 0)
        {
            file_ID = SDF_fopen_WR(fileName, fileSize);
            SDF_fclose();
        }

        if (file_ID == ERROR || SDF_find(fileName) == ERROR)
        {
            putstr("! Failed to create file: ");
            putstr(fileName);  putstr("\n");
        }

        break;
    }
    default:  break;

    }  // end switch
}


/*^***********************************************************************************************
*
*   CLI command function:  Cmnd_sst25
*
*   This command is for testing and debugging the SST25VF016B serial (SPI) data-flash
*   driver functions defined in module "../Board/SST25VF016.c".
*
*   The SST25VF016 serial data flash memory is 2MB in 512 sectors each of 4096 bytes.
*   Highest byte address is $1FF:FFF. 3 LS hex digits address a byte within a sector.
*/
void  Cmnd_sst25(int argCount, char *argVal[])
{
    static  uint32  lAddr = 0;          // Default write address = 0x000000
    static  uint32  dAddr = 0;          // Default dump address = 0x000000
    uint8   abPageBuf[256];             // Page buffer for 'read array' command
    char    cOption;
    uint16  device_id;

    if (!SuperUserAccess()) return;

    if ( argCount == 1 || argVal[1][0] == '?')
    {
        putstr( "Usage:  sst25  <opt> [addr] \n" );
        putstr( "<opt> \n" );
        putstr( " -c  +++   : Chip erase \n" );
        putstr( " -i        : Device ID  \n" );
        putstr( " -s        : Status reg \n" );
        putstr( " -d <addr> : Dump page (256 bytes) at <addr> \n" );
        putstr( " -e <addr> : Erase sector at <addr> (sss000) \n" );
        putstr( " -w <addr> : Write & verify sector at <addr> \n" );
    }

    cOption = tolower( argVal[1][1] );

    if ( argCount >= 3 && cOption == 'd') dAddr = long_hexatoi(argVal[2]) & 0x1FFF00;
    else if (argCount >= 3) lAddr = long_hexatoi(argVal[2]) & 0x1FF000;

    switch ( cOption )
    {
    case 'c':
        if (strmatch(argVal[2], "+++"))
        {
            SST25ResetWriteProtection();
            SST25ChipErase();
        }
        break;

    case 'i':
        device_id = SST25ReadDeviceID();
        putstr("Device ID = $");  putHexWord(device_id);
        putstr("\n");
        break;

    case 's':
        putstr("Status byte = $");  putHexByte(SST25ReadStatusReg());
        putstr("\n");
        break;

    case 'd':
        SST25ReadArray(dAddr, (uint8 *) abPageBuf, 256);
        putstr("Flash Addr: $");  putHexWord(dAddr >> 8);  putHexByte(dAddr);
        putstr("  Sector: $");  putHexByte(dAddr >> 16);  putHexDigit(dAddr >> 12);
        putstr("\n\n");
        dump_page_buffer( abPageBuf, dAddr );
        dAddr += 256;  // next page
        break;

    case 'e':
        SST25ResetWriteProtection();
        SST25SectorErase( lAddr );
        break;

    case 'w':
        SST25ResetWriteProtection();
        SST25SectorErase( lAddr );
        if (fill_flash_4K_block(lAddr) == FALSE)
            putstr("! Write/verify failed.\n");
        else  putstr("* Verified OK.\n");
        break;

    default:  break;

    }  // end switch
}


/*
*   Sub-function of Cmnd_sst25 --
*   Dumps a 256-byte page RAM buffer at a given address.
*   Displayed address is Flash address, not the RAM buffer address.
*/
static void  dump_page_buffer(uint8 *pbAddr, int iFlashAddr)
{
    char    c;
    uint8   iRow, iCol;

    for ( iRow = 0;  iRow < 16;  iRow++ )
    {
        putHexByte( (uint8) (iFlashAddr >> 16) );  // Show Flash address (3 bytes)
        putHexWord( (uint16) iFlashAddr );
        iFlashAddr += 16;
        putch(' ');
        for ( iCol = 0;  iCol < 16;  iCol++ )    // Hex dump row
        {
            if ( (iCol % 8) == 0 ) putch(' ');
            putHexByte( *pbAddr++ );
        }
        pbAddr -= 16;
        putch(' ');
        putch(' ');
        for ( iCol = 0;  iCol < 16;  iCol++ )    // ASCII dump row
        {
            c = *pbAddr++;
            if (c >= 0x20 && c < 0x7F)  putch( c );  // printable ASCII char
            else  putch('`');
        }
        putstr("\n");
    }
}


/*
*   Sub-function of Cmnd_sst25 --
*   Fills a 4K-byte block of SPI flash at lAddr with psuedo-random data.
*   Returns FALSE if a write/verify error occurred.
*/
static uint8  fill_flash_4K_block(uint32 lAddr)
{
    int  i;
    uint8  bdat, result = 1;

    srand(5551 + (lAddr >> 12));

    for (i = 0;  i < 4096;  i++, lAddr++)
    {
        bdat = rand() & 0xFF;
        SST25WriteByte(bdat, lAddr);
        if (SST25ReadByte(lAddr) != bdat) { result = 0;  break; }
    }

    return result;
}

#endif // USE_CONSOLE_CLI


#ifdef USE_MDD_FILE_SYSTEM
/**************************************************************************************************
 * Utility functions for data transfer between SDF memory and MDD media (USB drive or SD card).
 * ------------------------------------------------------------------------------------------------
 *
 * Function reads first/next record from INHX32 hex file on the MDD device
 * into a char buffer.
 *
 * Entry args:  inputBuf = pointer to record input buffer (max. 78 chars)
 *              inpFile  = file handle of input file on MDD media (already open)
 *
 * Return val:  length = number of printable chars in input record.
 */
int  ReadNextRecord(char *inputBuf, FSFILE *inpFile)
{
    int   length = 0;

    while (1)  // This loop looks for a start-of-record char (:)
    {
        if (FSfread(inputBuf, 1, 1, inpFile) == 0)  break;  // end of file

        if (*inputBuf == ':')  { length++;  inputBuf++;  break; }  // start of record
    }

    while (length < 79)  // This loop fetches remainder of record (if any)...
    {
        if (FSfread(inputBuf, 1, 1, inpFile) == 0)  break;  // end of file

        if (*inputBuf == '\r' || *inputBuf == '\n')  // end of record
            { *inputBuf = 0;  break; }

        length++;  inputBuf++;
    }

    return length;
}


/**
 * Function reads an INHX32 hex file on the MDD device to determine the target data
 * memory size occupied, based on the highest target address found in the input file,
 * i.e. in the last data record, and the number of data bytes in that record.
 * It is assumed that target memory space begins at absolute addr zero, i.e. there is
 * no relative offset.
 *
 * If the input Hex file contains a 'title' record (type = 0x09), the title string
 * is extracted (truncated to first 30 chars) and copied to a caller-supplied buffer.
 * The 'title record' is an Axxin proprietary extension to the INHX32 hex file format.
 * Title string format is "[xxxx...x]" where 'x' is any printable ASCII char except
 * SPACE, ':', '[' or ']'. The opening delimiter '[' follows the record type (09).
 * Title string length is limited to 30 chars (excluding delimiters).
 *
 * Entry args:  inpFile = file handle of input hex file (already open);
 *              title = pointer to buffer for title string extraction.
 *
 * Returns:     target memory size (bytes) occupied by input file data, or...
 *              ERROR (-1) if a problem with the input file format is found.
 *
 * INHX32 Hex record format: ":nnaaaattdd ... kk\r\n" (all digits are hex)
 * where:  "nn"   is the record size (data byte count),
 *         "aaaa" is the target address (LS 16 bits),
 *         "tt"   is the record type (00 = data, 01 = EOF, 04 = 'page' addr, 09 = title),
 *         "kk"   is the record checksum (2's complement of sum of bytes),
 *         "dd"   is the first data byte (if rec type is "00").
 */
static int  GetHexFileSize(FSFILE *inpFile, char *title)
{
//  char   printBuf[80];         // used for debug/trace only
    char   inHexRecordBuf[80];
    char  *pBuf = inHexRecordBuf;
    int    charCount;
    int    recordLen, recordAddr, recordType;
    int    recordCount = 0;      // number of DATA records in file
    int    previousAddr = 0;     // 32 bits
    int    absoluteAddr = 0;     // 32 bits
    int    highestAddr = 0;      // 32 bits
    int    bytesInLastDataRecord = 0;
    int    pageAddr = 0x0000;    // A 'page' in this context is 64kB (65536 bytes)

    while (1)
    {
        charCount = ReadNextRecord(inHexRecordBuf, inpFile);
        if (charCount == 0)  break;  // End of file reached before EOF record found

        sscanf(inHexRecordBuf, ":%2X%4X%2X", &recordLen, &recordAddr, &recordType);

        if (recordType == 0x09)  // have 'title' record...
        {
            pBuf = &inHexRecordBuf[10];
            while (*pBuf != ']')   // end of title string
            {
                if (*pBuf <= 0x20)  break;  // Ctrl char
                if ((pBuf - inHexRecordBuf) > 30)  break;  // too many chars
                *title++ = *pBuf++;
                *title = 0;  // Term char
            }
        }

        if (recordType == 0x00)  // have 'data' record...
        {
            recordCount++;
            absoluteAddr = (pageAddr << 16) | recordAddr;
            if (absoluteAddr < previousAddr)  // File is not sorted...
            {
/*              // <<<<<<<<<<<<<<<<<<<<<< debug/trace >>>>>>>>>>>>>>>>>>>>>>>>>>>>
                putstr("! Error: Input file not sorted...\n");
                sprintf(printBuf, "%d data records read.  Page addr = $%04X \n",
                        recordCount, pageAddr);
                putstr(printBuf);
                sprintf(printBuf, "Absolute Addr = $%06X,  Previous Addr = $%06X\n",
                        absoluteAddr, previousAddr);
                putstr(printBuf);
*/
                recordCount = 0;  // force ERROR exit
                break;
            }
            if (absoluteAddr > highestAddr)
            {
                highestAddr = absoluteAddr;
                bytesInLastDataRecord = recordLen;
            }
            previousAddr = absoluteAddr;
        }
        else if (recordType == 0x04)   // have 'page address' record...
        {
            sscanf(inHexRecordBuf, ":%2X%4X%2X%4x",
                   &recordLen, &recordAddr, &recordType, &pageAddr);
        }
        else if (recordType == 0x01)  break;  // have 'end-of-file' record
    }

    if (recordCount != 0)
        return (highestAddr + bytesInLastDataRecord);
    else  return (ERROR);
}


/**
 * Function reads data records from input hex file on MDD media (USB drive or SD card)
 * and writes the data to a file on SDF memory (SST25 device).
 *
 * The function requires that data records in the input file are sorted into ascending
 * address order, but the file may be fragmented.
 * If the input file is fragmented (most likely), gaps in the target address space
 * are filled with padding (value = 0xFF).
 *
 * Entry args:  inpFile = file handle of MDD media input hex file (already open);
 *              outFileSize = (max) number of bytes to be written to the output file.
 *
 * Return val:  byteCount = number of bytes written to the output file, or...
 *              ERROR (-1) if a checksum error or other data error is found.
 */
static int  ImportHexFileSub(FSFILE *inpFile, int outFileSize)
{
    char   inHexRecordBuf[80];
    char   endofFile = 0;
    uint8  dataByte, checksum;
    int    recordCount = 0;       // number of data records read
    int    byteCount = 0;
    int    recordLen, recordAddr, recordType, recordCksm, pos, i;
    int    targetAddr = 0;        // 32 bits
    int    absoluteAddr = 0;      // 32 bits
    int    pageAddr = 0x0000;

    while (byteCount <= outFileSize)
    {
        if (ReadNextRecord(inHexRecordBuf, inpFile) <= 0)  // Read failed
            { endofFile = TRUE;  break; }

        sscanf(inHexRecordBuf, ":%2X%4X%2X", &recordLen, &recordAddr, &recordType);

        if (recordType == 0x00)  // have data record...
        {
            recordCount++;
            checksum = (uint8) recordLen + (uint8) HI_BYTE(recordAddr) +
                       (uint8) LO_BYTE(recordAddr) + (uint8) recordType;
            absoluteAddr = (pageAddr << 16) | recordAddr;
            while (targetAddr < absoluteAddr)  // write pad byte(s) to output file
            {
                SDF_fwrite(0xFF);
                targetAddr++;
                byteCount++;
            }

            for (pos = 9, i = 0;  i < recordLen;  i++)  // write data bytes to output file
            {
                dataByte = hexctobin(inHexRecordBuf[pos]) << 4;  // MS digit
                dataByte += hexctobin(inHexRecordBuf[pos+1]);    // LS digit
                checksum += dataByte;
                SDF_fwrite(dataByte);
                targetAddr++;
                byteCount++;
                pos += 2;
            }
            recordCksm = hexctobin(inHexRecordBuf[pos]) << 4;  // checksum MS digit
            recordCksm += hexctobin(inHexRecordBuf[pos+1]);    // checksum LS digit

            if ((uint8)(0 - checksum) != recordCksm)  // Checksum error detected...
            {
/*
                // <<<<<<<<<<<<<<<<<<<<<<<<<< debug output >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
                putstr("Data record: ");  putDecimal(recordCount, 1);  putstr(",  ");
                putstr("Load addr: $");  putHexWord((uint16) recordAddr);  putstr(",  ");
                putstr("Calc cksm: $");  putHexByte((uint8)(0 - checksum));  putstr(",  ");
                putstr("Record cksm: $");  putHexByte(recordCksm);  putstr("\n");
*/
                byteCount = ERROR;  // Checksum error, force exit
                break;
            }
        }
        else if (recordType == 0x04)  // have 'page address' record
        {
            sscanf(inHexRecordBuf, ":%2X%4X%2X%4x",
                   &recordLen, &recordAddr, &recordType, &pageAddr);
        }
        else if (recordType == 0x01)  break;   // have 'end-of-file' record
    }

    return byteCount;
}


/**************************************************************************************************
 *
 * Function copies a file from SST25 SPI flash memory to MDD media (FAT32 file system).
 *
 * Called by Cmnd_sdf(), also by Cmnd_export, with '-b' option for binary copy.
 * If mode is binary (raw), the MDD output file size and data format is the same as the
 * SDF input file.
 *
 * Called by Cmnd_sdf(), also by Cmnd_export, with '-t' option for text copy.
 * If mode is text, copying terminates when an EOT marker (0xFF) is found.
 */
void  ExportFileFromSDF(int argCount, char *argVal[])
{
    DataFlash_FileIndexRecord_t *pFIR;

    short  file_ID;
    FSFILE *pFile;
    char   SDF_filename[24];    // up to 19 chars
    char   exportFileName[16];  // MSDOS '8.3' format
    uint8  sectorBuf[SST25_SECTOR_SIZE];
    int    i, iData, totalBytesCopied = 0;
    int    sectorsRemaining;    // # of SST25 sectors to be read
    int    byteCount;
    char   mode = 'B';          // default is Binary

    if (strmatch(argVal[1], "-t")) mode = 'T';

    if (!g_MDD_MediaMounted)
    {
        putstr("! USB drive not mounted.\n");
        return;
    }

    if (argCount != 4)
    {
        putstr("! Command syntax error.\n");
        return;
    }

    file_ID = atoi(argVal[2]);
    pFIR = SDF_finfo(file_ID);

    if (file_ID == 0 || pFIR == NULL)
    {
        putstr("! Invalid SDF file ID. \n");
        return;
    }

    strncpy(SDF_filename, pFIR->fileName, 20);
    SDF_filename[19] = 0;

    if (!SDF_fopen_RD(file_ID))
    {
        putstr("! Failed to open SDF input file: ");
        putstr(SDF_filename);
        putstr("\n");
        return;
    }

    sectorsRemaining = pFIR->linkSector - pFIR->begSector;

    // assume argVal[3] is MDD file-name (in 8.3 short format)
    strncpy(exportFileName, argVal[3], 13);
    exportFileName[12] = 0;
    pFile = FSfopen(exportFileName, "w");
    
    if (pFile == NULL)
    {
        putstr("! Failed to open output file: ");
        putstr(exportFileName);
        putstr("\n");
        return;
    }

    while (sectorsRemaining)
    {
        // Read SDF file sector (or part thereof if last) into buffer
        for ((byteCount = 0, i = 0); i < SST25_SECTOR_SIZE; i++)
        {
            iData = SDF_fread();
            if (iData < 0) break;    // end-of-file
            if (mode == 'T' && (iData & 0xFF) == 0xFF) break; // end-of-text marker
            ////
            sectorBuf[i] = (uint8) iData;
            byteCount++;
            totalBytesCopied++;
        }

        // Write (byteCount) bytes from buffer to MDD file
        if (FSfwrite(sectorBuf, byteCount, 1, pFile) != 1)
            break;
        sectorsRemaining--;
    }

    SDF_fclose();

    if (FSfclose(pFile) == ERROR)
        putstr("! Failed to close file: ");
    else
    {
        putDecimal(totalBytesCopied, 1);
        putstr(" bytes copied to file: ");
    }

    putstr(exportFileName);  putstr("\n");
    return;
}


/*^
 * Function reads Intel Hex (INHX32) file from MDD media (USB drive or SD card)
 * and creates a binary data file in SDF memory.
 *
 * Called by Cmnd_sdf(), also by Cmnd_import, both with '-x' option.
 * Command syntax:  sdf -x  <MDD_filename>  [SDF_filename]
 *
 * If <SDF_filename> is supplied as a command-line arg (argVal[3]), it will be used;
 * otherwise, if the input hex file contains a 'title' record (type = 09), the SDF
 * output filename will be extracted from the record (up to 19 printable chars);
 * otherwise the command will complain and exit without writing the SDF file.
 *
 * Data records in the input file must be sorted into ascending address order.
 * Data records must not contain more than 32 data bytes.
 */
void  ImportHexFileToSDF(int argCount, char *argVal[])
{
    DataFlash_FileIndexRecord_t *pFIR;

    short  file_ID;
    FSFILE *pFile;
    char   titleString[32];        // String extracted from Hex file 'title' record
    char   inputFileName[16];      // MDD media -- FAT32 '8.3' format
    char   outputFileName[24];     // SDF memory -- up to 19 chars
    int    fileSize, byteCount;
    uint8  gotOutputFilename = 0;

    if (!g_MDD_MediaMounted)
    {
        putstr("! MDD storage device not mounted.\n");
        return;
    }

    if (argCount < 3)
    {
        putstr("! Need input filename. \n");
        return;
    }
    strncpy(inputFileName, argVal[2], 13);  // assume argVal[2] is FAT32 file-name
    inputFileName[12] = 0;

    if (argCount == 4)  // assume argVal[3] is SDF memory file-name
    {
        strncpy(outputFileName, argVal[3], 20);
        outputFileName[19] = 0;
        gotOutputFilename = TRUE;
    }

    pFile = FSfopen(inputFileName, "r");
    if (pFile == NULL)
    {
        putstr("! Failed to open MDD media file: ");
        putstr(inputFileName);  putstr("\n");
        return;
    }

    fileSize = GetHexFileSize(pFile, titleString);
    if (fileSize == ERROR || fileSize == 0)
    {
        FSfclose(pFile);
        putstr("! Problem reading input hex file. (Normalized?) \n");
        return;
    }

    if (!gotOutputFilename && strlen(titleString) != 0)
    {
        // Use title string from input hex file as filename for SDF output file...
        strncpy(outputFileName, titleString, 20);  // limit 19 chars max.
        outputFileName[19] = 0;
        gotOutputFilename = TRUE;
    }

    if (!gotOutputFilename)
    {
        putstr("! Output file-name not available.\n");
        return;   // Game over
    }

    file_ID = SDF_fopen_WR(outputFileName, fileSize);

    if (file_ID == ERROR)
    {
        putstr("! Failed to open SDF memory file: ");
        putstr(outputFileName);  putstr("\n");
        return;
    }
    putstr("Writing to SDF file: ");  putstr(outputFileName);  putstr("\n");

    FSrewind(pFile);
    byteCount = ImportHexFileSub(pFile, fileSize);
    FSfclose(pFile);
    SDF_fclose();

    if (byteCount == fileSize)  putstr("File transfer complete. ");
    if (byteCount < fileSize)  putstr("! File transfer incomplete. ");
    if (byteCount == ERROR)  putstr("Checksum error detected.\n");
    else  { putDecimal(byteCount, 1);  putstr(" bytes written.\n"); }
}


/*^
 * Function copies a file from MDD media (FAT32 file system) to SST25 SPI flash memory.
 * The SDF output file format is the same as the MDD media input file.
 *
 * Called by Cmnd_sdf() with '-i' option ('i' for import);
 * also called by Cmnd_import with '-b' option ('b' for binary).
 *
 * Command arg's:  <cmd_name> <opt> <MDD filename> <SDF filename>
 *
 * Caller should check legality of the SDF filename.
 */
void  ImportBinaryFileToSDF(int argCount, char *argVal[])
{
    DataFlash_FileIndexRecord_t *pFIR;

    short  file_ID;
    FSFILE *pFile;
    char   inputFileName[16];      // MDD media -- FAT32 '8.3' format
    char   outputFileName[24];     // SDF memory -- up to 19 chars
    long   fileSize, byteCount = 0;
    uint8  bDat;

    if (!g_MDD_MediaMounted)
    {
        putstr("! MDD storage device not mounted.\n");
        return;
    }

    if (argCount < 4)
    {
        putstr("! Command error: Missing filename.\n");
        return;
    }

    strncpy(inputFileName, argVal[2], 13);  // assume argVal[2] is FAT32 file-name
    inputFileName[12] = 0;

    strncpy(outputFileName, argVal[3], 20);
    outputFileName[19] = 0;

    if (strlen(outputFileName) < 3)
    {
        putstr("! Output filename not valid.\n");
        return;
    }

    pFile = FSfopen(inputFileName, "r");
    if (pFile == NULL)
    {
        putstr("! Failed to open MDD input file: ");
        putstr(inputFileName);  putstr("\n");
        return;
    }

    fileSize = pFile->size;   // Input file size (bytes)
    file_ID = SDF_fopen_WR(outputFileName, fileSize);
    if (file_ID == ERROR)
    {
        putstr("! Failed to open SDF file for writing: ");
        putstr(outputFileName);  putstr("\n");
        FSfclose(pFile);
        return;
    }

    putstr("Writing to SDF file: ");
    putstr(outputFileName);  putstr(" ...\n");

    while (byteCount < fileSize)
    {
        if (FSfread(&bDat, 1, 1, pFile) == 0) break;  // End of file

        if (SDF_fwrite(bDat) == 0)  break;  // Write failed

        byteCount++;
    }

    FSfclose(pFile);
    SDF_fclose();

    if (byteCount == fileSize)  putstr("* Import complete. ");
    else  putstr("! Import failed. ");

    putDecimal(byteCount, 1);  putstr(" bytes written to SDF file.\n");
}


/*^
 * CLI command function:  Cmnd_export
 *
 * The "export" command is intended for transferring data from a file in SDF memory
 * to a file on the MDD mass storage device (USB drive or SD card).
 *
 * The MDD media output file format may or may not be the same as the SDF input file,
 * depending on the specified command option.
 *
 */
void  Cmnd_export(int argCount, char * argVal[])
{
    if (argCount == 1 || argVal[1][0] == '?')
    {
        putstr( "Usage:  export <opt> <SDF fileID#> <MDD filename> \n" );
        putstr( "  where <MDD filename> must use FAT32 '8.3' format\n" );
        putstr( "<opt> = -b : Binary mode -- copy verbatim\n" );
        putstr( "<opt> = -t : Text mode -- stop at EOT marker ($FF)\n" );
#ifdef USE_USB_MSD_HOST_INTERFACE
        putstr( "Transfer data from SDF memory file to USB drive.\n" );
#else
        putstr( "Transfer data from SDF memory file to SD card.\n" );
#endif
        return;
    }

    ExportFileFromSDF(argCount, argVal);
}


/*^
 * CLI command function:  Cmnd_import
 *
 * The "import" command is intended for transferring data from a file on the MDD mass
 * storage device (USB drive or SD card) to a file in SDF memory.
 *
 * The SDF output file format may or may not be the same as the MDD media input file,
 * depending on the specified command option.
 */
void  Cmnd_import(int argCount, char *argVal[])
{
    if (argCount == 1 || argVal[1][0] == '?')
    {
        putstr( "Usage:  import <opt> <MDD filename> <SDF filename>  \n" );
        putstr( "   where <MDD filename> must use FAT32 '8.3' format \n" );
        putstr( "   ... <SDF filename> is up to 19 chars (no spaces) \n" );
        putstr( "<opt> = -b : Binary mode -- copy verbatim \n" );
        putstr( "<opt> = -x : Read Hex (INHX32) format, write binary \n" );
#ifdef USE_USB_MSD_HOST_INTERFACE
        putstr( "Transfer data from USB drive to SDF memory file.\n" );
#else
        putstr( "Transfer data from SD card to SDF memory file.\n" );
#endif
        return;
    }

    if (strmatch(argVal[1], "-x"))   // Hex mode
        ImportHexFileToSDF(argCount, argVal);
    else  // Binary (raw) mode
        ImportBinaryFileToSDF(argCount, argVal);
}

#endif // USE_MDD_FILE_SYSTEM
