/**************************************************************************************************
 *
 * FileName:  data_flash_fs.h
 *
 * Note:      This version supports ST25VF016 device, 2MByte capacity.
 */
#ifndef DATA_FLASH_FS_H
#define DATA_FLASH_FS_H

#include "system_def.h"

#define SST25_TOTAL_SECTORS    512   // 2MB = 12 sectors x 4096 bytes
#define SST25_SECTOR_SIZE     4096   // bytes per sector

#define SDF_MEMORY_MAX_FILES  (SST25_SECTOR_SIZE / 32)   // Max. number of files in directory


// A data-flash 'File Index Record' (FIR) is created for each file in the volume.
// Each FIR occupies a 32-byte chunk of the File Directory sector(s).
//
typedef struct DataFlash_File_Index_Record  // FIR
{
    uint16  fileIndex;        // Index (file ID) of file in directory
    uint16  begSector;        // Sector address of first sector of this file
    uint16  linkSector;       // Sector address of first sector of next file
    uint16  reserved;         // (2 bytes reserved)
    uint32  fileSize;         // Number of data bytes in this file
    char    fileName[20];     // File name, incl. term-char (NUL)

} DataFlash_FileIndexRecord_t;


// Functions implemented in module "data_flash_fs.c"
//
uint8  SDF_format(void);                // Returns TRUE if the operation succeeded
int    SDF_fopen_WR(char *fileName, int32 fileSize);  // Returns file_ID if opened, or ERROR
uint8  SDF_fwrite(uint8 bDat);          // Returns TRUE if the operation succeeded
int    SDF_find(char *fileName);        // Returns file_ID if found, else ERROR (-1)
uint8  SDF_fopen_RD(uint16 fileID);     // Returns TRUE if the operation succeeded
int    SDF_fread(void);                 // Returns next unread byte, or End-Of-File (-1)
uint8  SDF_fclose(void);                // Returns TRUE if the operation succeeded
uint8  SDF_fhide(uint16 fileID);        // Returns TRUE if the operation succeeded
int    SDF_ferase(uint16 fileID);       // Returns ## of sectors left to be erased, or ERROR
uint8  SDF_freopen_WR(uint16 fileID);   // Returns TRUE if the operation succeeded

long   SDF_WriteArray(uint16 fileID, uint8 *pSourceBuf, long offset, long byteCount);
long   SDF_ReadArray(uint16 fileID, uint8 *pDestinBuf, long offset, long byteCount);

DataFlash_FileIndexRecord_t  *SDF_finfo(uint16 fileID);   // Returns pointer to FIR, or NULL

#if USE_CONSOLE_CLI
// CLI commands for managing files on the serial data-flash (SDF) memory device.
//
void  Cmnd_sdf(int argCount, char * argVal[]);
void  Cmnd_sst25(int argCount, char *argVal[]);
void  Cmnd_export(int argCount, char * argVal[]);
void  Cmnd_import(int argCount, char *argVal[]);
//
#endif

#ifdef USE_MDD_FILE_SYSTEM
// Functions for data transfer between SDF memory and MDD media (USB drive, SD card):
//
void  ExportFileFromSDF(int argCount, char *argVal[]);
void  ImportBinaryFileToSDF(int argCount, char *argVal[]);
void  ImportHexFileToSDF(int argCount, char *argVal[]);
int   ReadNextRecord(char *inputBuf, FSFILE *inpFile);  // Read from MDD file
//
#endif

#endif // DATA_FLASH_FS_H
