/* ================================================================================================
 *
 * FileName:    main_pic32_audio_test.c
 * `````````
 * Overview:    Main module for PIC32MX audio test application.
 * `````````
 * Processor:   PIC32MX340F512H  (See "HardwareProfile.h")
 * ``````````
 * Hardware:    Olimex PIC32-MX340 Prototyping Board (plus audio add-ons), or
 * `````````    Custom PIC32-MX340 Board (REMI Sound Synth Module)
 *
 * Compiler:    Microchip MPLAB XC32 v1.33++ (free version) under MPLAB'X IDE v2.20++.
 * `````````    The project contains no Microchip "PLIB" dependencies.
 *
 * Author:      M.J.Bauer, Copyright 2016-2018++, All rights reserved
 * ```````
 * Reference:   [1]  www.mjbauer.biz/Build_the_REMI_by_MJB.htm
 * ``````````   
 *
 * ================================================================================================
 */
#include <sys/attribs.h>    // For interrupt handlers

#include "main_pic32_audio_test.h"


PRIVATE  void   Init_Application(void); 
PRIVATE  void   StartupRoutine();
PRIVATE  void   MidiInputServiceTask();
PRIVATE  void   WaveTableSelect(uint8 wave_idx);


// -------------  Global data  --------------------------------------------------------
//
uint8   g_FW_version[4];           // firmware version # (major, minor, build, 0)
uint8   g_SelfTestFault[16];       // Self-test fault codes (0 => no fault)
uint32  g_counter;                 // temp, for debug
int32   g_TraceBuffer[100][5];
bool    g_RangeSwitchingTestActive;   // for "diag -r" command
int     g_RangeSwitchingPeriod_ms;

// These global variables may be modified by the CLI 'param' command...
float   g_WaveForm;                // Waveform ID code (0,1,2,3,...)
float   g_OscFrequency;            // Osc. frequency, Hz (up to 10000)
float   g_AudioLevel;              // Audio output level (0..100 %)
float   g_NoiseFilterFc;           // Noise filter corner freq, Hz (100..10000)


// --------------  Private data  ----------------------
//
static uint16   m_OscWaveTableSize;       // Number of samples in OSC wave-table
static float    m_OscFreqDiv;             // OSC frequency divider
static fixed_t  m_OscFreqMultiplier;      // Pitch-bend factor (range: 0.5 ~ 2.0)

static volatile bool     v_SynthEnable;   // Signal to enable sampling routine
static volatile int16   *v_WaveTable;     // address of active wave-table

static volatile fixed_t  v_OscAngle;      // sample position in wave-table
static volatile fixed_t  v_OscStep;       // sample position increment in table

static volatile uint16   v_NoiseFilterK;  // Noise filter time-constant
static volatile bool     v_NoiseEnable;   // Noise generator enable (0|1)
static volatile fixed_t  v_OutputLevel;   // Audio output level control (normalized)


//=================================================================================================

PRIVATE  void  Init_Application(void)
{
    int   i;

    for (i = 1;  i < NUMBER_OF_SELFTEST_ITEMS;  i++)
    {
        g_SelfTestFault[i] = 0;  // clear fault codes (except item 0)
    }

    g_FW_version[0] = BUILD_VER_MAJOR;
    g_FW_version[1] = BUILD_VER_MINOR;
    g_FW_version[2] = BUILD_VER_DEBUG;

    // Check that the MCU device ID matches the firmware build...
    if ((GET_DEVICE_ID() & DEVICE_ID_MASK) != DEVICE_ID_PIC32MX340F512H)
    {
        g_SelfTestFault[TEST_DEVICE_ID] = 1;
    }

    // Initialize PIC32 I/O pins, etc. (in pic32_low_level.c)
    Init_MCU_IO_ports();

    UART1_init(31250);   // for MIDI comm's
    Init_I2C1();         // for EEPROM
    
    g_AppTitleCLI = "PIC32MX Audio Synth Board Test Utility...\n";
    
    // Set default sound parameter values...
    g_WaveForm = SINE_WAVE;
    g_OscFrequency = 1000;      // Hz
    g_NoiseFilterFc = 20000;    // Fc = 40000/Kt Hz 
    g_AudioLevel = 100;         // %
}


int  main(void)
{
    int  status = InitializeMCUclock();

    UART2_init(57600);   // for console using UART2
    putstr("\n");
    putstr("* MCU reset/startup \n");
    putstr("* Running self-test routine \n");

    if (status == ERROR)
    {
        g_SelfTestFault[TEST_SOFTWARE_TIMER] = 1;
        putstr("! Timing error in Delay10us(): ");
        putDecimal(g_SoftTimerError, 1);
        putstr("% \n");
    }
    else  g_SelfTestFault[TEST_SOFTWARE_TIMER] = 0;

    Init_Application();

    if (g_SelfTestFault[TEST_DEVICE_ID])
        putstr("! PIC32 device type is incompatible with firmware build.\n");

    Cmnd_ver(1, NULL);          // Send startup message to console
    PrepareForNewCommand();     // Initialize CLI

    SynthInit();                // Initialize audio hardware and ISR

    while ((1 + 1) == 2)        // main loop
    {
        BackgroundTaskExec();

        ConsoleCLI_Service();
    }
}


/*
 * Background task executive...  Runs periodic tasks scheduled by the RTI timer ISR.
 * Also dispatches any pending asynchronous (non-periodic event-driven) tasks.
 *
 * This routine is called frequently from the main loop and from inside wait loops;
 * e.g.  while waiting for a timer to expire (see 'WaitMilliseconds()').
 * Foreground (interrupting) tasks are executed directly from their respective ISR's.
 */
void  BackgroundTaskExec()
{
    static  uint8  flag_10ms_task = 0;
    static  int    rangeSwitchTimer = 0;
    
    
//  if (g_DiagnosticModeActive)  return;  // Background tasks suspended
    
    if (v_RTI_flag_1ms_task)  // Do 1ms periodic tasks
    {
        // Call 1ms task function(s) here (none as yet)
		//
        v_RTI_flag_1ms_task = 0;
    }

    if (v_RTI_flag_5ms_task)  // Do 5ms periodic tasks
    {
//      ButtonInputService();

        flag_10ms_task = flag_10ms_task ^ 1;
        if (flag_10ms_task)    // Do 10ms periodic tasks
        {
            ;;;
        }
		//
        v_RTI_flag_5ms_task = 0;
    }

    if (v_RTI_flag_50ms_task)  // Do 50ms periodic tasks
    {
        // Call 50ms task function(s) here (none as yet)
        //
        v_RTI_flag_50ms_task = 0;
    }

    if (v_RTI_flag_500ms_task)  // Do 500ms periodic tasks
    {
        // Call 500ms task function(s) here (none as yet)
		//
        v_RTI_flag_500ms_task = 0;
    }
}


/*
 * Foreground task executive...
 * Runs periodic interrupting tasks sync'd to the RTI timer.
 * The function is called by the RTI Timer ISR on every RTI "tick" (1ms).
 */
void  ForegroundTaskExec()
{
    ;
}

//=================================================================================================

/*
 * Function:  SynthInit()
 * 
 * Initialize Timer 2 and OC4 pin for PWM audio output generation.
 * Set initial values for user-settable sound parameters.
 *
 * Timer_2 is used for the wave-table oscillators' PWM signal output.
 * Timer_2 is set up with Prescaler = 1:1;  Fclk = FCY = 80MHz;  Tclk = 12.5ns
 * For sampling rate = 40kHz...
 * Timer_2 period := 25.00us (2000 x 12.5ns);  hence PR2 = 1999.
 * Maximum duty register value is 1999.
 * Configure OC4 module for PWM, fault detect disabled. 
 * 
 */
void  SynthInit(void)
{
    TRISDbits.TRISD3 = 0;    // RD3/OC4 is an output pin
    OC4CON = 0x0000;         // Disable OC4 while timer is set up

    // Set up Timer 2 for 40kHz audio sample rate
    T2CON = 0;
    T2CONbits.TCKPS = 0;     // Prescaler set to 1:1
    PR2 = 1999;              // Period = 2000 x 12.5ns (freq = 40,000 Hz)

    IFS0bits.T2IF = 0;       // Clear IRQ flag
    IPC2bits.T2IP = 6;       // Set IRQ priority
    T2CONbits.TON = 1;       // Start Timer

    OC4R = 1000;             // PWM Set initial duty (50%)
    OC4RS = 1000;
    OC4CON = 0x8006;         // Enable OC4 for PWM

    v_SynthEnable = 0;
    WaveTableSelect(0);   // default wave-table (SINE_WAVE)
    SetAudioOutputLevel(IntToFixedPt(1) / 2);  // 50%

    TIMER2_IRQ_ENABLE();
}


/*
 * Function:     Select (activate) a wave-table for the synth oscillator.
 *
 * Entry args:   wave_idx = index of flash-based wave-table descriptor
 */
PRIVATE  void  WaveTableSelect(uint8 wave_idx)
{
    if (wave_idx < GetNumberOfFlashWaveTables())  // bounds check
    {
        m_OscWaveTableSize = g_FlashWaveTableDef[wave_idx].Size;
        m_OscFreqDiv = g_FlashWaveTableDef[wave_idx].FreqDiv;
        v_WaveTable = g_FlashWaveTableDef[wave_idx].Address;
    }
}


/*
 * Function:     Activate the wave-table oscillator at a given frequency.
 *
 * Overview:     Low-level function intended for audio hardware testing purposes.
 *               The oscillator is stopped by a call to SynthOscStop().
 */
void   SynthOscStart()
{
    float  oscFreq;
    int    level_pc = (int) g_AudioLevel;
    
    WaveTableSelect((int) g_WaveForm);  // Not used for noise gen.

    if ((int) g_WaveForm == NOISE_WAVE)  v_NoiseEnable = 1;
    else  v_NoiseEnable = 0;
    
    oscFreq = g_OscFrequency / m_OscFreqDiv;  // Apply OSC freq. divider
    v_OscAngle = 0;
    v_OscStep = FloatToFixed((oscFreq * m_OscWaveTableSize) / SAMPLE_RATE_HZ);
    v_NoiseFilterK = 40000 / g_NoiseFilterFc;   // filter time-constant (1..40000)
    
    // Set audio output level using square-law curve
    v_OutputLevel = IntToFixedPt(level_pc * level_pc) / 10000;     
            
    v_SynthEnable = 1;     // Let 'er rip, Boris.
}


/*
 * Function:     De-activate the synth oscillator. Mute audio output.
 *
 * Entry args:   --
 */
void  SynthOscStop()
{
    v_SynthEnable = 0;
    Delay10us(6);
}


/*`````````````````````````````````````````````````````````````````````````````````````````````````
 * Function:     Timer_2 interrupt service routine (ISR)
 *
 * The ISR performs audio synth computations which need to be executed at the PCM audio
 * sampling rate, defined by SAMPLE_RATE_KHZ (typically 40kHz). This routine implements
 * the wave-table oscillator, a noise generator, and a low-pass IIR noise filter.
 *
 * If SAMPLE_RATE_KHZ == 40, the periodic IRQ interval is 25us.  The ISR execution time
 * is approx 5us (measured).  With Fclk = 80 MHz, 1us equals 80 instruction cycles.
 */
void  __ISR(_TIMER_2_VECTOR, IPL6AUTO)  Timer_2_IRQService(void)
{
    static  int32   filterAccum = 0;      // Noise filter accumulator
    static  int32   filterOutput = 0;     // last sample from noise filter
    static  uint32  rand_last = 1;        // random noise seed (must be odd!)

    int32    oscWaveIdx;                  // index into wave-table
    int16    oscSample;                   // sample from wave-table osc.
    int32    noiseSample;                 // sample from noise generator (32b)
    int32    filterInput;                 // sample to be filtered
    int16    outputSample;                // sample to be output (16 bit)
    fixed_t  scaledSample;                // sample scaled by audio output level

    TESTPOINT_T2ISR_SET_HI();

    if (v_SynthEnable)
    {
        // Wave-table oscillator implementation...
        // Fetch sample from wave-table according to Osc. frequency
        oscWaveIdx = IntegerPart(v_OscAngle);
        oscSample = v_WaveTable[oscWaveIdx];
        v_OscAngle = (v_OscAngle + v_OscStep) % IntToFixedPt(m_OscWaveTableSize);

        // White noise generator -- Pseudo-random number algorithm...
        rand_last = (rand_last * 1103515245 + 12345) & 0x7FFFFFFF;
        noiseSample = (int16) ((uint16)(rand_last >> 15));  // signed 16-bit
        
        // Noise filter (1st-order low-pass IIR)
        filterInput = noiseSample >> 1;  // scale to 50% FS
        filterAccum -= filterOutput;  
        filterAccum += filterInput;
        filterOutput = filterAccum / v_NoiseFilterK;
         
        if (v_NoiseEnable)  outputSample = filterOutput;
        else  outputSample = oscSample;
    }
    else  // Sound muted
    {
        outputSample = 0;
    }
    
    // The normalized sample value is scaled by the ampld-controller output level,
    // giving a signed 15 bit integer in the range +/-16k.
    outputSample = outputSample >> 1;   // +/-0.999 FS  (sign + 14 LS bits)
    scaledSample = MultiplyFixed(outputSample, v_OutputLevel);
    
    // Scale the 15-bit sample to 11 bits (+/-1000 FS) and add 50% bias (1000) to get 
    // PWM duty register value in the range 0..1999 (positive 11 bit number).
    OC4RS = (scaledSample >> 4) + 1000;
    
    TESTPOINT_T2ISR_SET_LO();
    IFS0bits.T2IF = 0;         // Clear the IRQ
}


/*
 * Function:  Timer_3 interrupt service routine (ISR)
 *
 * Overview:  The ISR handles any task(s) which must be executed
 *            synchronously with the Timer_3 rollover (period).
 */
void __ISR(_TIMER_3_VECTOR, IPL5AUTO)  Timer_3_IRQService(void)
{
    IFS0bits.T3IF = 0;
}


/*
 * Function:     Set wave-table size parameter for wave-table used by OSC.
 *
 * Entry arg:    (uint16) size = OSC wave-table size (number of samples used).
 */
void  WaveTableSizeSet(uint16 size)
{
    m_OscWaveTableSize = size;
}


/*
 * Function:     Set OSC Frequency Divider value.
 *
 * Entry arg:    (float) freqDiv = Frequency Divider value
 */
void  OscFreqDividerSet(float freqDiv)
{
    m_OscFreqDiv = freqDiv;
}


/*
 * Function:     Get current Frequency Divider value used in OSC wave-table.
 *
 * Return val:   (float) freqDiv = Frequency Divider value
 */
float  OscFreqDividerGet()
{
    return  m_OscFreqDiv;
}


/*
 * Function:  Return TRUE if the Remi synth is enabled, i.e. if a note is in progress.
 */
bool  isSynthActive(void)
{
    return  (v_SynthEnable != 0);
}


/*
 * Function:     Set Audio Output Level.
 *
 * Entry args:   level = (fixed_t) audio output level, per unit (0 ~ 0.9999).
 *               Wave samples from the synth tone-generator are scaled by this quantity.
 * 
 * Note:         Diagnostic mode must be active for this function to be effective.
 */
void  SetAudioOutputLevel(fixed_t level_pu)
{
    v_OutputLevel = level_pu;
}
