/**************************************************************************************************
 *
 * FileName:        main_sound_fx_synth_app.c
 *
 * Processor:       PIC32MX460F512L
 *
 * H/W Platform:    Olimex PIC32-MX460 board, modified...
 *                  + SST25VFxxx SPI flash memory on SPI1 (CS# = RB2)
 *                  + MCP79410 real-time clock on I2C2 (optional)
 *
 * Compiler:        Microchip MPLAB XC32 v1.30++ (free version)
 *
 * Author:          M.J.Bauer, 2015   [www.mjbauer.biz]
 *
 * Overview:
 * `````````
 *     This PIC32 firmware app implements a PCM digital audio synthesizer and player
 *     designed to create short monophonic sound FX clips for use in embedded systems.
 *     An external serial (SPI) data-flash memory is used to store sound FX files.
 *     Sound FX files can be exported to and imported from a USB flash drive.
 *     The project provides a framework for other PIC32MX embedded applications.
 *
 * Reference:       "Digital Sound Effects for Embedded Systems (PIC32 MX460)", M.J.Bauer (2015)
 *
 * History:         See ../doc/RevisionHistory.txt
 *
 * ------------------------------------------------------------------------------------------------
 * Build Notes:
 * ````````````
 *     The MDD file system requires a heap in data RAM... 4k bytes recommended.
 *     (Project Properties / XC32 (Global Options) / xc32-ld / Heap size ...)
 *
 * ================================================================================================
 */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "system_def.h"
#include "main_sound_fx_synth_app.h"
#include "console.h"
#include "data_flash_fs.h"
#include "pcm_audio.h"

#define SYS_FREQ 	(80000000L)


uint8 g_FW_version[4];    // firmware version # (major, minor, build, 0)

//=================================================================================================

static void  Init_Application(void)
{
    static const DRV_SPI_INIT_DATA  SST25_SPI_config =
        { SST25_SPI_CHANNEL, SST25_SPI_SPIBRG, 0, 0, 1, 1, 0 };

    uint32  time_ms;
    short   count;

    g_FW_version[0] = BUILD_VER_MAJOR;
    g_FW_version[1] = BUILD_VER_MINOR;
    g_FW_version[2] = BUILD_VER_DEBUG;

    // Initialize I/O, periph's. (in low-level.c)
    Init_MCU_IO_ports();

    UART2_init(57600);   // for console using UART2
    Init_I2C2();         // for RTCC, ext I/O, etc

#ifdef CHECK_SOFTWARE_TIMER_ACCURACY  // Build option -- see main_xxxx.h

    time_ms = millisecTimer();
    for (count = 0;  count < 1000;  count++)
    {
        Delay10us(10);  // delay 100us  (in TimeDelay.c)
    }
    time_ms = millisecTimer() - time_ms;  // elapsed time (ms)

    if (time_ms < 90 || time_ms > 110)  // should be 100ms (+/-10%)
    {
        putstr("\n! Timing error in Delay10us(): ");
        putDecimal(((((int)time_ms - 100) * 100)/100), 1);
        putstr("% \n");
    }

#endif // CHECK_SOFTWARE_TIMER_ACCURACY

    // Initialize the SPI flash (SST25VFxxx) interface;
    SST25Init((DRV_SPI_INIT_DATA *) &SST25_SPI_config);

    SDF_fclose();    // Initialize the SDF file system

    srand(555 + TMR1);
    PCM_Audio_Init();
    Sound_Synth_Init();
}


int main(void)
{
    InitializeMCUclock();
    Init_Application();
    USBInitialize(0);

    putstr("\n");
    putstr("PIC32MX Sound FX Synthesizer \n");
    Cmnd_ver(1, NULL);
    PrepareForNewCommand();

    RTCC_Synchronize();  // Set time & date from hardware RTCC

    PlaySound(short_beep_2500Hz, 1600);

    while (1)
    {
        BackgroundTaskExec();
        ConsoleCLI_Service();
        USB_Service();
    }
}


/**
 * Function:     SystemHeartbeatLEDtask()
 *
 * Overview:     Pulses the "system heartbeat" LED at 2Hz (100ms duty).
 *               Periodic B/G task called at 50ms intervals.
 */
void SystemHeartbeatLEDtask(void)
{
    static short LED_cycle = 0;  // unit = 50ms

    if (LED_cycle >= 10) LED_cycle = 0;
    if (LED_cycle == 0) DebugLEDControl(1);
    if (LED_cycle == 2) DebugLEDControl(0);
    LED_cycle++;
}


/**
 * Background task executive...  Runs periodic tasks scheduled by the RTI timer ISR
 * and any asynchronous ("ad-hoc", non-periodic, event-driven) tasks.
 *
 * This routine is called frequently from the main loop and from inside all wait loops;
 * e.g.  while waiting for a timer to expire (see 'wait_milliseconds').
 * Foreground (interrupting) tasks are executed directly from their respective ISR's.
 */
void BackgroundTaskExec(void)
{
    // Asynchronous ("ad-hoc", non-periodic) tasks may be called here...
    //

    if (v_RTI_flag_5ms_task)
    {
        // Function(s) to be executed every 5 mSec are called here...
        //
        v_RTI_flag_5ms_task = 0;
    }
    if (v_RTI_flag_50ms_task)
    {
        // Function(s) to be executed every 50 mSec are called here...
        SystemHeartbeatLEDtask();
        //
        v_RTI_flag_50ms_task = 0;
    }
    if (v_RTI_flag_500ms_task)
    {
        // Function(s) to be executed every 500 mSec are called here...
#ifdef USE_MDD_FILE_SYSTEM
        MDDmediaMountStatusUpdate();
        UpdateFSrealTimeVars();
#endif
        //
        v_RTI_flag_500ms_task = 0;
    }
    if (v_RTI_flag_NewDayRollover)
    {
        RTC_NewDayRolloverTask();
        v_RTI_flag_NewDayRollover = 0;
    }
}


/**
 * Foreground task executive...
 * Runs interrupting tasks sync'd to the RTI timer.
 * The function is called by the RTI Timer ISR on every RTC "tick" (typ. 1ms).
 */
void  ForegroundTaskExec(void)
{
#ifdef USE_4_WIRE_KEYPAD
    ADC_Service_Keypad();
#endif
}


/*```````````````````````````````````````````````````````````````````````````````````````
 * CLI command function:  Cmnd_eeprom
 *
 * Examine or erase EEPROM contents.
 *
 * Note: The I2C EEPROM (24LC08) is not used in this application framework, but this
 *       command function demonstrates usage of the EEPROM driver functions.
 */
void  Cmnd_eeprom(int argCount, char * argValue[])
{
    uint8   blockNum = 0;
    uint8   buffer[20];
    uint8   eebyte;
    int     offset, row, col, count;
    char    option = tolower(argValue[1][1]);

//  if (!SuperUserAccess()) return;

    if (argCount == 1 || *argValue[1] == '?' )  // help
    {
        putstr( "Usage:  eeprom  <opt>  <block> \n" );
        putstr( "where <opt> = \n" );
        putstr( "  -d : Dump  EEPROM block (0..3) \n" );
        putstr( "  -e : Erase EEPROM block (0..3) \n" );
        return;
    }

    if (EepromReadData(buffer, 0, 0, 16) != 16)
    {
        putstr("! Error: EEPROM read access failed. \n");
        return;
    }

    for (count = 0;  count < 16;  count++)
    {
        buffer[count] = 0xFF;
    }

    if (argCount > 2)  blockNum = atoi(argValue[2]) & 3;

    if (option == 'd')  // Dump block contents
    {
        putstr("EEPROM block #");
        putHexDigit(blockNum);
        putstr(" = \n");

        for (offset = 0, row = 0;  row < 16;  row++, offset += 16)
	{
            EepromReadData((uint8 *) buffer, blockNum, offset, 16);
            putHexByte(offset);
            putch(':');

            for (col = 0;  col < 16;  col++)        // Hex dump
            {
                if ((col % 4) == 0)  putch(' ');
                putHexByte(buffer[col]);
            }
            putch(' ');

            for (col = 0;  col < 16;  col++)        // ASCII dump
            {
                eebyte = buffer[col];
                if (isprint(eebyte))  putch(eebyte);
                else  putch('.');
            }
            putstr("\n");
	}
    }

    if (option == 'e')  // Erase block contents
    {
        putstr("EEPROM block #");
        putHexDigit(blockNum);

        for (offset = 0, count = 0;  count < 256;  count += 16)
        {
            if (EepromWriteData(buffer, blockNum, offset, 16) == ERROR)
                break;
            offset += 16;
        }

        if (count == 256) putstr(" erased OK.\n");
        else  putstr(" erase failed.\n");
    }
}


/**
*   Function called by "default" command function...
*/
void  DefaultPersistentData(void)
{
    ;;;
}

/**
*   Function called by "commit" command function...
*/
void  CommitPersistentParams(void)
{
    ;;;
}

/**
*   Function called by "diag" command function...
*/
void  DiagnosticCommandExec(int argCount, char * argValue[])
{
    ;;;
}

/**
*   Function called by "watch" command function...
*   Variables to be "watched" are output on a single line (no newline).
*/
void  WatchCommandExec(void)
{
    ;;;
}

