/*=======================================================================================
 *
 * Module:      pic32_audio_test_cli.c
 *
 * Overview:    CLI command functions specific to the PIC32 Audio Test application.
 *
 *=======================================================================================
 */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>

#include "console_cli.h"
#include "pic32_audio_test_cli.h"
#include "main_pic32_audio_test.h"


extern  int32   g_TraceBuffer[][5];
extern  float   g_WaveForm;                // Waveform ID code (0,1,2,3,4...)
extern  float   g_OscFrequency;            // Osc. frequency, Hz (max. 10,000)
extern  float   g_AudioLevel;              // Audio output level (0..100 %)
extern  float   g_NoiseFilterFc;           // Noise filter corner freq, Hz (100..10,000)

char    *g_AppTitleCLI;                    // Title string output by "ver" command

/*----------------------------------------------------------------------------------------
*                  A P P L I C A T I O N   C O M M A N D   T A B L E
*/
const  CmndTableEntry_t  AppCommands[] =
{
    //   Cmd Name      Attribute      Cmd Function
    //----------     -------------   -------------------
    {    "eeprom",     DBG_CMD,       Cmnd_eeprom     },
    {    "info",       APP_CMD,       Cmnd_info       },
    {    "sound",      APP_CMD,       Cmnd_sound      },
    {    "trace",      APP_CMD,       Cmnd_trace      },
    {    "util",       APP_CMD,       Cmnd_util       },
    
    //---------------------------------------------------
    {    "$",          0,             NULL            }   // Dummy last entry
} ;


// Table of user-settable parameters, for use with the "set" command (alias "param").
// These parameters are stored in data RAM, must be declared as global and type float.
//
const  UserSettableParameter_t  UserParam[] =
{
    // nickname      disp  (float *) &g_VarName      min, max
    { "wave",        'i',  &g_WaveForm,               0,  9     },
    { "freq",        'r',  &g_OscFrequency,       0.001,  10000 },
    { "level",       'i',  &g_AudioLevel,             0,  100   },
    { "noiseFc",     'i',  &g_NoiseFilterFc,          4,  40000 },
    { "$",           'r',  NULL,                      0,  0     }  // end of table
};


/*```````````
 * Synthesizer sound test, executed by "sound" or "diag" command.
 *
 * Entry arg gives the sound duration in seconds. Value zero -> continuous sound.
 * Waveform and osc. frequency are given by global parameters via the "set" command.
 */
PRIVATE  void  SynthSoundTest(int duration)
{
    static  uint32  testEndTime;

    testEndTime = millisecTimer() + duration * 1000;  // msec

    SynthOscStart();

    if (duration > 0)  // run for the specified time
    {
        while (millisecTimer() < testEndTime)
        {
            BackgroundTaskExec();
        }

        SynthOscStop();
    }
}


/*```````````````````````````````````````````````````````````````````````````````````````
 * CLI command function:  Cmnd_eeprom
 *
 * Examine or erase EEPROM contents.
 */
void  Cmnd_eeprom(int argCount, char * argValue[])
{
    uint8   blockNum = 0;
    uint8   buffer[20];
    uint8   eebyte;
    int     offset, row, col, count;
    char    option = tolower(argValue[1][1]);

//  if (!SuperUserAccess()) return;

    if (argCount == 1 || *argValue[1] == '?' )  // help
    {
        putstr( "Usage:  eeprom  <opt>  <block> \n" );
        putstr( "where <opt> = \n" );
        putstr( "  -d : Dump  EEPROM block (0..3) \n" );
        putstr( "  -e : Erase EEPROM block (0..3) \n" );
        return;
    }

    if (EepromReadData(buffer, 0, 0, 16) != 16)
    {
        putstr("! Error: EEPROM read access failed. \n");
        return;
    }

    for (count = 0;  count < 16;  count++)
    {
        buffer[count] = 0xFF;
    }

    if (argCount > 2)  blockNum = atoi(argValue[2]) & 3;

    if (option == 'd')  // Dump block contents
    {
        putstr("EEPROM block #");
        putHexDigit(blockNum);
        putstr(" = \n");

        for (offset = 0, row = 0;  row < 16;  row++, offset += 16)
        {
            EepromReadData((uint8 *) buffer, blockNum, offset, 16);
            putHexByte(offset);
            putch(':');

            for (col = 0;  col < 16;  col++)        // Hex dump
            {
                if ((col % 4) == 0)  putch(' ');
                putHexByte(buffer[col]);
            }
            putch(' ');

            for (col = 0;  col < 16;  col++)        // ASCII dump
            {
                eebyte = buffer[col];
                if (isprint(eebyte))  putch(eebyte);
                else  putch('.');
            }
            putstr("\n");
        }
    }

    if (option == 'e')  // Erase block contents
    {
        putstr("EEPROM block #");
        putHexDigit(blockNum);

        for (offset = 0, count = 0;  count < 256;  count += 16)
        {
            if (EepromWriteData(buffer, blockNum, offset, 16) == ERROR)
                break;
            offset += 16;
        }

        if (count == 256) putstr(" erased OK.\n");
        else  putstr(" erase failed.\n");
    }
}


/*
 *   CLI command function:  Cmnd_info
 *
 *   The "info" command shows "hard-wired" system configuration info.
 */
void  Cmnd_info(int argCount, char * argValue[])
{
    char   textBuf[100];
    uint8  HWconfig = GetHardwareConfig();

    sprintf(textBuf, "PIC32 Audio Test firmware:  v%d.%d.%02d \n",
            g_FW_version[0], g_FW_version[1], g_FW_version[2]);
    putstr(textBuf);

    sprintf(textBuf, "HW config. jumper setting:  %d \n", HWconfig);
    putstr(textBuf);
}


/*
 *   CLI command function:  Cmnd_sound
 *
 *   The "sound" command initiates a sound generated by the synth engine.
 */
void  Cmnd_sound(int argCount, char * argValue[])
{
    int   runtime = atoi(argValue[2]);
    
    if (argCount == 1 || *argValue[1] == '?' )  // help
    {
        putstr( "Usage:  sound  {1 | 0}  [arg] \n" );
        putstr( "   where:  1 = On, 0 = Off, <arg> = duration (sec) \n" );
        putstr( "If no arg given, or arg <= 0, sound is continuous.  \n" );
        putstr( "-----\n" );
        putstr( "Waveform, Frequency, Audio Level, Noise filter Fc, etc, \n" );
        putstr( "are adjusted using the 'set' command, where... \n" );
        putstr( "   wave = 0:noise, 1:sine, 2:triangle, 3:square, 4:sawtooth \n" );
        putstr( "   freq = 0.001 ~ 10000 Hz \n" );
        putstr( "   level = 0 ~ 100 % (square-law curve) \n" );
        putstr( "   noiseFc = 20 ~ 20000 Hz \n" );
        putstr( "To avoid aliasing in non-sinusoidal waveforms, set freq to...  \n" );
        putstr( "  20, 25, 40, 50, 80, 100, 200, 250, 400, 500, 1k, 2k, 2.5k, 5k \n" );
        return;
    }
    
    if (*argValue[1] == '1')  SynthSoundTest(runtime);
    else  SynthOscStop(); 
}


/*```````````````````````````````````````````````````````````````````````````````````````
 *   Function called by "diag" system command.
 *
 *   NB:  All diag command options (except '-x') activate Diagnostic Mode
 *        which causes the synth process (B/G task) to be suspended.
 *        Command "diag -x" cancels Diagnostic Mode and re-activates the synth
 *        process. A time-out (default 30 mins) is imposed on Diagnostic Mode.
 */
void  DiagnosticCommandExec(int argCount, char * argValue[])
{
    static bool  diagEnabled;
    char   option = tolower(argValue[1][1]);
    char   opt_c2 = tolower(argValue[1][2]);

    if ( argCount == 1 || *argValue[1] == '?' )   // help wanted
    {
        putstr( "REMI low-level diagnostics \n" );
        putstr( "Usage:  diag  <option>  [arg] \n" );
        putstr( "Options:  \n" );
        putstr( " -s  :  Sound test (arg = duration, sec) \n");
        putstr( " -b  :  Read PORTB pins \n");
        putstr( " -t  :  Timeout for diag mode (mins) \n");
        putstr( " -x  :  Exit diagnostic mode. \n");
        ////
        return;
    }

    if (option != 'x')
    {
        if (!diagEnabled)
        {
            putstr("! Some background tasks may be suspended...\n");
            putstr("  Use 'diag -x' to resume normal operation.\n");
        }
        
        EnableDiagnosticMode();
        diagEnabled = 1;
    }

    switch (option)
    {
    case 's':  // Sound test using wave-table osc (or noise gen.)
    {
        int   runtime = atoi(argValue[2]);    // sound duration (sec)

        if (argCount < 3)  runtime = 0;  // sound continuously

        SynthSoundTest(runtime);
        break;
    }
    case 'b':  // Read PORTB pins
    {
        putstr("PortB pins: $");
        putHexWord(PORTB);
        NEW_LINE;
        break;
    }
    case 't':  // Set timeout for diagnostic mode;  arg = timeout (mins)
    {
        int   arg = atoi(argValue[2]);  // minutes

        if (arg != 0)
        {
            g_DiagnosticModeTimeout = arg;
            EnableDiagnosticMode();
        }
        break;
    }
    case 'x':  // Exit (cancel) diagnostic mode
    {
        CancelDiagnosticMode();
        diagEnabled = 0;
        putstr("* Normal synth operation resumed.\n");
        
        g_RangeSwitchingTestActive = 0;
        break;
    }
    default:  break;
    } // end switch
}


/*`````````````````````````````````````````````````````````````````````````````````````````````````
 *   CLI command function:  Cmnd_trace
 *
 *   The "trace" command outputs the debug trace buffer.
 *   Contents of the trace buffer depends on application-specific debug code.
 */
void  Cmnd_trace(int argCount, char *argValue[])
{
    char   textBuf[100];
    int    line;
    float  var1, var2, var3, var4;

//  if (!SuperUserAccess()) return;

    if (*argValue[1] == '?')   // help wanted
    {
        putstr( "Usage:  trace \n" );
        putstr( "Dumps the trace buffer. \n" );
        ////
        return;
    }

    putstr("   T ms    final  smoothed  level(13b) \n");

    for (line = 0;  line < 100;  line++)
    {
        var1 = FixedToFloat(g_TraceBuffer[line][1]);  // final ampld
        var2 = FixedToFloat(g_TraceBuffer[line][2]);  // smoothed ampld

        sprintf(textBuf, "%7d  %7.3f  %7.3f  %7d \n",
                g_TraceBuffer[line][0], var1, var2, g_TraceBuffer[line][3]);
        putstr(textBuf);
    }
}


/*
 *   CLI command function:  Cmnd_util
 *
 *   The "util" command executes a specified utility.
 */
void  Cmnd_util(int argCount, char *argValue[])
{
    char   option = tolower(argValue[1][1]);

    if (!SuperUserAccess()) return;

    if (argCount == 1 || *argValue[1] == '?')   // help wanted
    {
        putstr( "Run a specified Utility...   \n" );
        putstr( "Usage:  util  <opt1>  [arg] \n" );
        putstr( "<opt1> \n" );
        putstr( "  -a : ... \n");
        putstr( "  -b : ... \n");
        putstr( "  -c : ... \n");
         ////
        return;
    }

    switch (option)
    {
	default: break;
    }
}


/*```````````````````````````````````````````````````````````````````````````````````````
 *   Function called by "watch" command function...
 *   Variables to be "watched" are output on a single line (no newline).
 */
void  WatchCommandExec(void)
{
    putstr("  Press [Esc]  ");
}


/*```````````````````````````````````````````````````````````````````````````````````````
 *   Function called by "default" command function...
 *   Restores "factory default" values to persistent data (in EEPROM).
 */
void  DefaultPersistentData(void)
{
//  DefaultConfigData();
//  DefaultPresetData();
}


/*```````````````````````````````````````````````````````````````````````````````````````
 *   Function called by "commit" command function...
 *   Saves "working values" (RAM variables) of user-settable parameters to
 *   non-volatile memory (EEPROM), so that they become persistent.
 */
void  CommitPersistentParams()
{
    ;;;
}
