/*^
 * Module:       sound_fx_synth.c
 * ```````
 * Overview:     PCM digital audio sound effects synthesizer.
 * `````````
 * Originated:   M.J.Bauer, 2014   [www.mjbauer.biz]
 * ```````````
 * This module provides functions to synthesize sound effects for embedded applications.
 * Sounds can be played using the PCM audio player in module "pcm_audio.c".
 *
 * Attention:    This module is not intended for inclusion in the "REMI" controller project;
 * ``````````    it is incompatible with the "REMI" controller application.
 *
 * Various sound effects can be synthesized according to a number of preset parameters.
 * Sample data can be output in C source code syntax for the declaration of an initialized
 * array of (short) integers, for use in other applications.
 * In addition, sample data can be output to a USB flash drive or other mass-storage.
 *
 * Reference:  "Digital Sound Effects for Embedded Systems (PIC32_MX460)", M.J.Bauer
 * ``````````
 * Notes:
 * ``````
 *    Patch 1: Ring modulation (Quadrature AM, output = OSC1 * OSC2)
 *    Patch 2: Frequency modulation (OSC1 freq := OSC2, 5kHz Full-Scale)
 *    Patch 3: Patch 1 + OSC1 frequency sweep (OSC1_FREQ param ignored)
 *
 *    In Patch 1 or 3, to disable OSC2, set OSC2_LEVEL = 0.0, OSC2_OFFSET = 1.0.
 *    In Patch 1 or 3, QAM becomes simple AM if OSC2 output is always positive.
 *    Control AM or FM modulation depth using OSC2_LEVEL and OSC2_OFFSET.
 *
 *    The osc#level parameter can be used to invert the polarity of the waveform;
 *    e.g. set osc1level = -1.0 to make OSC1 RAMP_WAVE ramp down, +1.0 to ramp up.
 *
 * ================================================================================================
 */

#include "pcm_audio.h"
#include "sound_fx_synth.h"
#include "data_flash_fs.h"
#include "console.h"

#include <math.h>

// Set the maximum clip length == sample buffer file size (words).
// 100,000 samples => 5 seconds playing time @ 20k samples/sec.
//
#define MAX_CLIP_LENGTH    100000    // number of samples (words)

// Supported wave shapes for the oscillators --
#define  SINE_WAVE       1
#define  TRIANG_WAVE     2
#define  SQUARE_WAVE     3
#define  RAMP_WAVE       4
#define  PRNG_WAVE       5
#define  NOISE           6

#ifndef PI
#define PI  3.141593
#endif

enum Envelope_phases
{
    ENV_HOLDOFF,
    ENV_ATTACK,
    ENV_SUSTAIN,
    ENV_RELEASE
};


static float OSC1WaveTable[1000];     // 1 cycle of OSC1 periodic waveform
static float OSC2WaveTable[1000];     // 1 cycle of OSC2 periodic waveform

// Synthesizer user-settable parameters (adjusted using "set" command)...
//
static float duration;     // ms
static float patch;        // 1..3
static float osc1wave;     // 1..4
static float osc1freq;     // Hz
static float osc1level;    // 0..+/-1.0
static float osc1offset;   // 0..+/-1.0
static float osc2wave;     // 1..4
static float osc2freq;     // Hz
static float osc2level;    // 0..+/-1.0
static float osc2offset;   // 0..+/-1.0
static float holdoff;      // ms
static float attack;       // ms
static float sustain;      // ms
static float release;      // ms
static float sweepbegf;    // Hz
static float sweependf;    // Hz
static float sweeptime;    // ms
static float sweephold;    // boolean flag
static float prngseed;     // 0..65535
static float prngfbck;     // 0..65535


const  UserSettableParameter_t  UserParam[] =    // For use with "param" command
    {
        { "duration",  'i',  &duration,   0, 5000 },
        { "patch",     'i',  &patch,      1, 10 },
        { "osc1wave",  'i',  &osc1wave,   1, 6 },
        { "osc2wave",  'i',  &osc2wave,   1, 6 },
        { "osc1freq",  'r',  &osc1freq,   0.01, 5000.01 },
        { "osc2freq",  'r',  &osc2freq,   0.01, 5000.01 },
        { "osc1level", 'r',  &osc1level,  -1.0, 1.0 },
        { "osc2level", 'r',  &osc2level,  -1.0, 1.0 },
        { "osc1ofst",  'r',  &osc1offset, -1.0, 1.0 },
        { "osc2ofst",  'r',  &osc2offset, -1.0, 1.0 },
        { "holdoff",   'i',  &holdoff,    0, 5000 },
        { "attack",    'i',  &attack,     0, 5000 },
        { "sustain",   'i',  &sustain,    0, 5000 },
        { "release",   'i',  &release,    0, 5000 },
        { "sweepbegf", 'i',  &sweepbegf,  0, 5000 },
        { "sweependf", 'i',  &sweependf,  0, 5000 },
        { "sweeptime", 'i',  &sweeptime,  0, 5000 },
        { "sweephold", 'i',  &sweephold,  0, 1.1 },
        { "prngseed",  'i',  &prngseed,   -99999, 99999 },
        { "prngfbck",  'i',  &prngfbck,   0, 99999 },
        { "volume",    'i',  &g_AudioVolume, 0, 100 },
        //-----------------------------------------------------
        { "$",         'r',  NULL, 0.0, 0.0 }  // Last entry!
    };


static short fidLayerBuffer;     // SDF file ID # of current layer buffer
static short fidSwapBufferPing;  // SDF file ID # of output buffer "Ping"
static short fidSwapBufferPong;  // SDF file ID # of output buffer "Pong"
static short fidOutputBuffer;    // SDF file ID # of active output buffer
static short fidMergeBuffer;     // SDF file ID # of shadow output buffer

static short m_Layer;            // sound editor layer
static int32 clipLength;         // number of samples @ 20ks/s
static BOOL  isOutputBufferEmpty;
static BOOL  isBufferFileError;

// ----------  private functions  --------------------
//
static void  GenerateLayer(void);
static void  PopulateWavetables(void);
static int16 PseudoRandomInt16(int16 seed, uint16 fbmask);
static void  EraseLayerBuffer(void);
static void  MergeLayerIntoOutputBuffer(void);
static int   FindCompositePeakAmplitude(void);
static void  SaveSoundToMDDmediaRawFile(int argCount, char *argVal[]);
static void  SaveSoundToMDDmediaHexFile(int argCount, char *argVal[]);
static void  SetDefaultParamValues(void);
static void  ListCurrentLayerParams(FSFILE *stream);
static void  DumpSoundBufferAs_C_sourceArray(int argCount, char *argVal[]);

//=================================================================================================
/**
 *   Initialize the synthesizer environment
 */
void  Sound_Synth_Init(void)
{
    uint32  bufferSize = MAX_CLIP_LENGTH * 2;  // bytes

    SetDefaultParamValues();

    fidLayerBuffer = SDF_find("layer-buffer.pcm");
    if (fidLayerBuffer == ERROR)  // file not found -- create it
    {
        fidLayerBuffer = SDF_fopen_WR("layer-buffer.pcm", bufferSize);
        SDF_fclose();
    }

    fidSwapBufferPing = SDF_find("swap-buf-ping.pcm");
    if (fidSwapBufferPing == ERROR)  // file not found -- create it
    {
        fidSwapBufferPing = SDF_fopen_WR("swap-buf-ping.pcm", bufferSize);
        SDF_fclose();
    }

    fidSwapBufferPong = SDF_find("swap-buf-pong.pcm");
    if (fidSwapBufferPong == ERROR)  // file not found -- create it
    {
        fidSwapBufferPong = SDF_fopen_WR("swap-buf-pong.pcm", bufferSize);
        SDF_fclose();
    }

    if (fidLayerBuffer == ERROR || fidSwapBufferPing == ERROR || fidSwapBufferPong == ERROR)
    {
        isBufferFileError = TRUE;
        return;
    }
    else  isBufferFileError = FALSE;

    fidOutputBuffer = fidSwapBufferPing;
    fidMergeBuffer = fidSwapBufferPong;

    isOutputBufferEmpty = 1;
    m_Layer = 1;
    GenerateLayer();           // Generate layer 1 default sound
}


/**
 *   CLI command function:  Cmnd_synth
 *
 *   The "synth" command generates PCM audio sample data for a wave-table file.
 *   Sample values are signed 12-bit integers, scaled to cover the range +/-2000.
 *
 */
void  Cmnd_synth( int argCount, char * argVal[] )
{
    char   option;
    int    file_ID;
    short  fidSwapping;

    if (argCount == 1 || *argVal[1] == '?' )   // help wanted
    {
        putstr( "Usage:  synth  <option>  [arg] \n" );
        putstr( "<option>-----------------------------------------\n" );
        putstr( "  -l :  List layer parameters (verbose) \n" );
        putstr( "  -n :  New sound, current layer = 1  \n" );
        putstr( "  -g :  Generate sample data in current layer    \n" );
        putstr( "  -m :  Merge current layer into output buffer   \n" );
        putstr( "  -u :  Undo last merge  \n" );
        putstr( "  -t :  Test sound in current layer \n" );
        putstr( "  -p :  Play merged sound in output buffer       \n" );
        putstr( "  -s :  Save clip to USB drive as RAW file...    \n" );
        putstr( "          <arg> = clip_name (optional) \n" );
        putstr( "          MDD file-name generated automatically  \n" );
        putstr( "  -x :  Save clip to USB drive as HEX file...    \n" );
        putstr( "          <arg> = clip_name (up to 19 chars),    \n" );
        putstr( "          MDD file-name generated automatically  \n" );
        putstr( "  -d :  Dump sound clip buffer as C array def'n  \n" );
        putstr( "          <arg> = clip_name (optional) \n" );
        putstr( "  -v :  View values for patch, osc#wave, etc.    \n" );
        putstr( "-------------------------------------------------\n" );
        return;
    }

    if (argVal[1][0] == '-') option = argVal[1][1];
    else option = '$';

    if (isBufferFileError)
    {
        putstr("! Failed to initialize synth buffer file(s).\n");
        return;
    }

    clipLength = duration * 20;   // number of samples @ 20ks/s

    switch (option)
    {
    case 'n':  // New sound
        m_Layer = 1;
        isOutputBufferEmpty = TRUE;
        putstr("Output buffer is void... Now editing Layer 1.\n");
        break;

    case 'g':
        GenerateLayer();
        PlaySoundFile(fidLayerBuffer, clipLength);
        break;

    case 'm': 
        MergeLayerIntoOutputBuffer();
        PlaySoundFile(fidOutputBuffer, clipLength);
        break;

    case 'u':  // Undo merge -- swap merge buffer with output buffer
        if (m_Layer > 1)
        {
            fidSwapping = fidMergeBuffer;
            fidMergeBuffer = fidOutputBuffer;
            fidOutputBuffer = fidSwapping;
            if (--m_Layer == 1) isOutputBufferEmpty == TRUE;
            putstr("* Undo OK -- Now editing Layer ");
            putDecimal(m_Layer, 1);  putstr("\n");
        }
        else  putstr("! Can't undo -- Layer 1 not merged.\n");
        break;

    case 't':  // test the sound stored in the layer buffer file
        PlaySoundFile(fidLayerBuffer, clipLength);
        break;

    case 'o':  // alias for 'p' ...
    case 'p':  // play the sound stored in the output buffer file
        if (isOutputBufferEmpty)
        {
            putstr("! Output buffer is void... Playing Layer 1.\n");
            PlaySoundFile(fidLayerBuffer, clipLength);
        }
        else  PlaySoundFile(fidOutputBuffer, clipLength);
        break;

    case 's':  // Save output buffer to SDF memory file (raw binary)
        SaveSoundToMDDmediaRawFile(argCount, argVal);
        break;

    case 'x':  // Save output buffer to MDD file in Intel Hex format (INHX32)
        SaveSoundToMDDmediaHexFile(argCount, argVal);
        break;

    case 'l':  // List layer param's -- output to console
        putstr("Clip length: ");  putDecimal(clipLength, 1);
        putstr(" samples, duration: ");
        putDecimal((int) duration, 1);  putstr(" ms\n");
        ListCurrentLayerParams(NULL);
        break;

    case 'd':  // Dump output buffer to console as C source file
        DumpSoundBufferAs_C_sourceArray(argCount, argVal);
        break;

    case 'v':  // Show param values for patch number, osc wave-shape, etc,
        putstr("patch = 1:QAM, 2:FM, 3:SWEEP \n");
        putstr("osc#wave = 1:Sine, 2:Triangle, 3:Square, 4:Ramp, 5:PRNG, 6:Noise^ \n");
        putstr("^Noise is for OSC2 only (osc2freq ignored).\n");
        break;

    default:
        putstr("! Undefined option: -");  putch(option);  NEW_LINE;
        break;
        
    } // end switch
}


/**
 *   CLI command function:  Cmnd_play
 *
 *   The "play" command starts playback of a PCM sound clip stored in SDF memory.
 *   Arg1 is a sound file ID number (not zero).
 */
void  Cmnd_play(int argCount, char *argVal[])
{
    DataFlash_FileIndexRecord_t  *fileInfo;
    int  file_ID;

    if (argCount >= 2)  // file_ID supplied...
    {
        file_ID = atoi(argVal[1]);

        if (file_ID > 0)
        {
            fileInfo = SDF_finfo(file_ID);
            PlaySoundFile(file_ID, (fileInfo->fileSize / 2));
        }
    }
    else
    {
        putstr( "Usage:  play  <file_ID> \n" );
        return;
    }
}


// Return the duration (in msec) of the sound clip being synthesized.
//
uint32  GetSoundClipDuration(void)
{
    return  duration;  // milliseconds
}


//---------------------------------------------------------------------------------------
// Output current layer param's to a given stream (stdout or MDD media file).
//
static void ListCurrentLayerParams(FSFILE *stream)
{
    char   outBuf[100];

    sprintf(outBuf, "Layer %d parameters...\n", m_Layer);
    fputstr(outBuf, stream);

    strcpy(outBuf, "PATCH : ");
    if (patch == 1) strcat(outBuf, "(1) Ring Modulator (output = OSC1 * OSC2) \n");
    else if (patch == 2) strcat(outBuf, "(2) Frequency Modulator (OSC2 -> OSC1 freq) \n");
    else if (patch == 3) strcat(outBuf, "(3) Patch 1 + Sweep OSC1 frequency \n");
    fputstr(outBuf, stream);

    strcpy(outBuf, "OSC 1 : ");
    if (osc1wave == SQUARE_WAVE) strcat(outBuf, "Square (3)");
    else if (osc1wave == SINE_WAVE) strcat(outBuf, "Sine (1)");
    else if (osc1wave == TRIANG_WAVE) strcat(outBuf, "Triangle (2)");
    else if (osc1wave == RAMP_WAVE) strcat(outBuf, "Sawtooth (4)");
    else if (osc1wave == PRNG_WAVE) strcat(outBuf, "Random (5)");
    fputstr(outBuf, stream);

    if (patch == 1)  // OSC1 freq is fixed
        sprintf(outBuf, ", freq= %4d Hz, level= %5.3f, offset= %5.3f \n",
                (int) osc1freq, osc1level, osc1offset);
    else  // OSC1 freq. is modulated by OSC2, or controlled by SWEEP gen.
        sprintf(outBuf, ", level= %5.3f, offset= %5.3f \n", osc1level, osc1offset);
    fputstr(outBuf, stream);

    strcpy(outBuf, "OSC 2 : ");
    if (osc2wave == SQUARE_WAVE) strcat(outBuf, "Square (3)");
    else if (osc2wave == SINE_WAVE) strcat(outBuf, "Sine (1)");
    else if (osc2wave == TRIANG_WAVE) strcat(outBuf, "Triangle (2)");
    else if (osc2wave == RAMP_WAVE) strcat(outBuf, "Sawtooth (4)");
    else if (osc2wave == PRNG_WAVE) strcat(outBuf, "Random (5)");
    else if (osc2wave == NOISE) strcat(outBuf, "Noise (6)");
    fputstr(outBuf, stream);

    if (osc2wave != NOISE)
    {
        sprintf(outBuf, ", freq= %4d Hz", (int) osc2freq);
        fputstr(outBuf, stream);
    }
    sprintf(outBuf, ", level= %5.3f, offset= %5.3f \n", osc2level, osc2offset);
    fputstr(outBuf, stream);

    if (osc1wave == PRNG_WAVE || osc2wave == PRNG_WAVE)
    {
        sprintf(outBuf, "PRNG  : seed = %d (0x%04X), feedback = %d (0x%04X) \n",
                (int) prngseed, (int) prngseed, (int) prngfbck, (int) prngfbck);
        fputstr(outBuf, stream);
    }

    sprintf(outBuf,
            "ENV   : holdoff= %d ms, attack= %d ms, sustain= %d ms, release= %d ms\n",
            (int) holdoff, (int) attack, (int) sustain, (int) release);
    fputstr(outBuf, stream);

    if (patch == 3)  // SWEEP on OSC1
    {
        sprintf(outBuf,
                "SWEEP : beg.freq= %d Hz, end.freq= %d Hz, time= %d ms, %s\n",
                (int) sweepbegf, (int) sweependf, (int) sweeptime,
                ((int) sweephold ? "HOLD" : "Repeat"));
        fputstr(outBuf, stream);
    }
}


//---------------------------------------------------------------------------------------
// This function generates PCM audio samples comprising the current sound layer.
// Samples are written to the layer buffer (file on SDF memory).
//
static void  GenerateLayer(void)
{
    char   textBuf[100];
    int    isample;             // output sample count
    int16  sampleValue;         // output sample value (signed 16-bit)
    float  envelope = 0.0;      // envelope magnitude (max: 1.0)
    float  osc1out, osc2out;    // oscillator output signals (normalized)
    float  modout;              // modulator output signal
    int    osc1angle = 0;       // wave-table angle, OSC1 (unit = 0.01*2*pi)
    int    osc2angle = 0;       // wave-table angle, OSC2 (range 0..99999)
    int    osc1frad, osc2frad;  // oscillator radian frequencies (angular increment)
    int    osc1idx, osc2idx;    // wave-table indexes (range 0..999)
    short  env_phase = ENV_HOLDOFF;
    int    delay_count = 0;
    int    sus_count = 0;
    int    sweep_count = 0;
    float  sweep_range = sweependf - sweepbegf;  // Hz
    float  osc2outMax = -1000.0, osc2outMin = 1000.0, osc2outSum = 0;
    uint16 fback = (uint16)(prngfbck + 0.001);  // round to nearest integer

    clipLength = duration * 20;                     // samples @ 20ks/s
    if (clipLength > 100000) clipLength = 100000;   // max 5 seconds @ 20 ks/s
    if (patch == 3) osc1freq = sweepbegf;           // OSC1 in sweep mode

    if (osc1wave == NOISE) osc1wave = PRNG_WAVE;    // OSC1 can't be NOISE!

    PseudoRandomInt16((int16) prngseed, fback);     // Seed the PRNG
    PopulateWavetables();
    EraseLayerBuffer();

    if (!SDF_freopen_WR(fidLayerBuffer))
    {
        putstr("! Failed to open layer buffer file.\n"); 
        return;
    }

    for (isample = 0; isample < clipLength; isample++)
    {
        if (++sweep_count >= ((int) sweeptime * 20))  // Sweep generator
        {
            if (sweephold) sweep_count = (int) sweeptime * 20; // sweep hold
            else sweep_count = 0;  // sweep repeat
        }

        if (osc2wave != NOISE)
        {
            // Fetch next OSC2 sample from wave-table according to OSC2 frequency
            osc2frad = (((int) osc2freq) << 8) / 20;  // = (osc2freq * 256) / 20
            osc2idx = osc2angle >> 8;  // get LUT index (0..999) from wave angle
            osc2angle += osc2frad;     // bump the wave angle by frequency
            osc2angle = osc2angle % 256000;   // 0..256000 is 360 degrees (2*pi radians)
            osc2out = OSC2WaveTable[osc2idx] * osc2level + osc2offset;
        }
        else  // osc2wave == NOISE
        {
            osc2out = ((float) ((int16) (rand() & 0xFFFF))) / 32768;  // white noise
//          osc2out = ((float) PseudoRandomInt16(0, fback)) / 32768;  // PRNG noise
            osc2out = osc2out * osc2level + osc2offset;
        }

        osc2outSum += osc2out;
        if (osc2out < osc2outMin) osc2outMin = osc2out;
        if (osc2out > osc2outMax) osc2outMax = osc2out;

        // todo: if (osc2wave != SINE) pass osc2out thru anti-alias filter (Fc = 5kHz)

        if (patch == 2)   // FM
            osc1freq = 50 + 3000 * osc2out;
        if (patch == 3)   // SWEEP
            osc1freq = sweepbegf + (sweep_range * sweep_count) / (sweeptime * 20);

        // Fetch next OSC1 sample from wave-table according to OSC1 frequency
        osc1frad = (((int) osc1freq) << 8) / 20;  // = (osc1freq * 256) / 20
        osc1idx = osc1angle >> 8;  // get LUT index (0..999) from wave angle
        osc1angle += osc1frad;     // bump the wave angle by frequency
        osc1angle = osc1angle % 256000;   // 0..256000 is 360 degrees (2*pi radians)
        osc1out = OSC1WaveTable[osc1idx] * osc1level + osc1offset;
        
        if (patch == 2) modout = osc1out;  // Freq modulator (FM)
        else  modout = osc1out * osc2out;  // Ring modulator (QAM)

        // todo: Pass modout thru anti-alias filter (Fc = 5kHz)

        sampleValue = modout * envelope * 30000;      // Apply ENV and scale sample to +/-30000
        if (sampleValue > 30000) sampleValue = 30000;    // Apply limiter (clipper)
        if (sampleValue < -30000) sampleValue = -30000;

        if (!SDF_fwrite(sampleValue & 0xFF))  // write LSB of sample
        {
            putstr("! Failed to write layer buffer file.\n");
            break;
        }

        SDF_fwrite(sampleValue >> 8);  // write MSB of sample

        // Compute next envelope magnitude value...
        if (env_phase == ENV_HOLDOFF)
        {
            // The 'holdoff' phase delays the envelope attack for a specified time...
            if (++delay_count >= (int) holdoff * 20) env_phase = ENV_ATTACK;
        }
        else if (env_phase == ENV_ATTACK)  
        {
            // The 'attack' phase is a linear ramp ending with envelope == 1.0 ...
            envelope = envelope + 1.0 / ((float) attack * 20.0);
            if (isample >= (holdoff + attack) * 20) env_phase = ENV_SUSTAIN;
        }
        else if (env_phase == ENV_SUSTAIN)
        {
            // The 'sustain' phase holds the envelope magnitude at its peak value...
            if (++sus_count >= (int) sustain * 20) 
            {
//              putDecimal(sus_count, 1);  putstr("\n");   // ** temp debug
                env_phase = ENV_RELEASE;
            }
        }
        else if (env_phase == ENV_RELEASE)
        {
            // The 'release' phase is an exponential decay from the peak value...
            envelope = envelope - (envelope / (release * 20.0));
        }
    }

    SDF_fclose();
/*
    // Diagnostic output
    sprintf(textBuf, "osc2out(min) = %6.3f \n", osc2outMin);
    putstr(textBuf);
    sprintf(textBuf, "osc2out(max) = %6.3f \n", osc2outMax);
    putstr(textBuf);
    sprintf(textBuf, "osc2out(ave) = %6.3f \n", osc2outSum / clipLength);
    putstr(textBuf);
*/
}


//---------------------------------------------------------------------------------------
// Function merges the current layer buffer with the output buffer (if not empty)
// and writes the composite sound data to a new output buffer.
// The composite sound samples are "normalized", i.e. scaled to maximize amplitude
// and to eliminate clipping.
//
static void  MergeLayerIntoOutputBuffer(void)
{
    int16 layerBuff[1000];
    int16 mergeBuff[1000];
    char  textBuf[80];
    int   max_sectors, tries;
    int   mcount = 0;   // samples merged
    int   i, chunk;
    long  offset = 0;
    short fidSwapping;
    int   sampleValue, peakAmpld;

    // Erase the Merge Buffer file (which becomes the Output Buffer after the merge)
    max_sectors = ((MAX_CLIP_LENGTH * 2) / SST25_SECTOR_SIZE) + 1;
    tries = max_sectors + 2;

    while (tries-- != 0)
    {
        if (SDF_ferase(fidMergeBuffer) == 0) break;  // done
    }

    clipLength = (int) duration * 20;  // sample count
    
    // To normalize the samples, find the highest composite sample value...
    peakAmpld = FindCompositePeakAmplitude();

    while (mcount < clipLength)
    {
        // Read SDF layer file and SDF output file into respective RAM buffers...
        SDF_ReadArray(fidLayerBuffer, (uint8 *) layerBuff, offset, 2000);  // 1000 samples
        if (!isOutputBufferEmpty)
            SDF_ReadArray(fidOutputBuffer, (uint8 *) mergeBuff, offset, 2000);

        // Merge and write 1000 samples to RAM merge buffer
        for (i = 0;  i < 1000;  i++)
        {
            if (isOutputBufferEmpty) sampleValue = 0;
            else sampleValue = mergeBuff[i];  // sample from SDF Output file

            sampleValue += layerBuff[i];  // merge samples
            sampleValue = (sampleValue * 30000) / peakAmpld;  // normalize result
            mergeBuff[i] = sampleValue;
        }

        if ((clipLength - mcount) >= 1000) chunk = 1000;
        else chunk = clipLength - mcount;

        // Flush RAM buffer to SDF merge file (up to 1000 samples)
        SDF_WriteArray(fidMergeBuffer, (uint8 *) mergeBuff, offset, chunk * 2);
        offset += 2000;     // bytes, not samples
        mcount += chunk;    // samples
    }

    // The Merge Buffer just written becomes the new Output Buffer...
    fidSwapping = fidOutputBuffer;
    fidOutputBuffer = fidMergeBuffer;
    fidMergeBuffer = fidSwapping;
    isOutputBufferEmpty = FALSE;

//  sprintf(textBuf, "%d samples merged.\n", mcount);
//  putstr(textBuf);
    sprintf(textBuf, "Layer %d added to buffer... ", m_Layer);
    putstr(textBuf);
    m_Layer++;
    sprintf(textBuf, "Now editing Layer %d.\n", m_Layer);
    putstr(textBuf);
}


// Function returns the highest magnitude composite sample which would result from
// merging (adding) current layer samples with output buffer samples.
//
static int  FindCompositePeakAmplitude(void)
{
    int16 layerBuff[1000];
    int16 outputBuff[1000];
    int   i;
    int   mcount = 0;   // samples processed
    long  offset = 0;
    int32 sampleAmpld;
    int32 peakAmpld = 0;  // peak sample amplitude

    clipLength = (int) duration * 20;  // sample count

    while (mcount < clipLength)
    {
        SDF_ReadArray(fidLayerBuffer, (uint8 *) layerBuff, offset, 2000);
        if (!isOutputBufferEmpty)
            SDF_ReadArray(fidOutputBuffer, (uint8 *) outputBuff, offset, 2000);
        offset += 2000;  // bytes (not samples)
        mcount += 1000;  // samples

        for (i = 0;  i < 1000;  i++)
        {
            if (isOutputBufferEmpty) outputBuff[i] = 0;
            sampleAmpld = abs(outputBuff[i] + layerBuff[i]);
            if (sampleAmpld > peakAmpld) peakAmpld = sampleAmpld;
        }
    }

    if (peakAmpld == 0) peakAmpld = 30000;  // return something usable

    return  peakAmpld;
}


//---------------------------------------------------------------------------------------
// This function populates the two oscillator wave-tables each with one cycle of a
// periodic waveform which may be sinusoidal, square, triangle, ramp (saw-tooth),
// or a pseudo-random sequence (RANDOM_WAVE).
// Wave-table values are bipolar floats, normalized to peak amplitude = 1.0.
//
static void  PopulateWavetables(void)
{
    float  sample;
    int    index;
    uint16 fback = (uint16)(prngfbck + 0.001);  // round to nearest integer

    for (index = 0; index < 1000; index++)
    {
        sample = sinf((2 * PI) * (float) index / 1000);  // SINE wave
        
        if (osc1wave == SINE_WAVE) OSC1WaveTable[index] = sample;
        if (osc2wave == SINE_WAVE) OSC2WaveTable[index] = sample;
        
        sample = (index < 500) ? 1.0 : -1.0;  // SQUARE wave
        
        if (osc1wave == SQUARE_WAVE) OSC1WaveTable[index] = sample;
        if (osc2wave == SQUARE_WAVE) OSC2WaveTable[index] = sample;
            
        if (index < 250) sample = 1.0 * ((float) index / 250);  // TRIANGE wave
        else if (index < 500) sample = 1.0 - (float) (index - 250) / 250;
        else if (index < 750) sample = -1.0 * (float) (index - 500) / 250;
        else sample = (float) (index - 750) / 250 - 1.0;
        
        if (osc1wave == TRIANG_WAVE) OSC1WaveTable[index] = sample;
        if (osc2wave == TRIANG_WAVE) OSC2WaveTable[index] = sample;
        
        sample = (2.0 * (float) index) / 1000 - 1.0;  // SAWTOOTH wave
        
        if (osc1wave == RAMP_WAVE) OSC1WaveTable[index] = sample;
        if (osc2wave == RAMP_WAVE) OSC2WaveTable[index] = sample;

        sample = ((float) PseudoRandomInt16(0, fback)) / 32768;  // PRNG wave

        if (osc1wave == PRNG_WAVE) OSC1WaveTable[index] = sample;
        if (osc2wave == PRNG_WAVE) OSC2WaveTable[index] = sample;
    }
}


/*
 * PRNG -- Pseudo-Ramdom Number Generator for PRNG wave table stuffing
 *
 * Entry arg's: (uint16) seed   = initial bit pattern if non-zero (1..65535)
 *                                ... no effect if zero.
 *              (uint16) fbmask = feedback bit mask (0x0000..0xFFFF)
 *
 * Return val:  (int16) Next number in sequence
 */
static int16  PseudoRandomInt16(int16 seed, uint16 fbmask)
{
    static  uint32  PRNG_register;  // using 16 LS bits for result
    int     i;
    uint16  exorbits;
    uint16  si = 1;  // shift-reg serial input
    
    if (seed != 0) 
    {
        PRNG_register = ((uint32) seed) & 0xFFFF;
    }
    else
    {
        exorbits = PRNG_register & fbmask;

        for (i = 0;  i < 16;  i++)
        {
            si = si ^ (exorbits & 0x0001);
            exorbits = exorbits >> 1;
        }

        if (si == 1) PRNG_register |= 0x10000;  // Set bit16
        PRNG_register = PRNG_register >> 1;
    }

    // Truncate to 16 bits and convert to signed integer
    return (int16)((uint16) PRNG_register);
}


// Erase the current layer buffer
//
static void  EraseLayerBuffer(void)
{
    int  max_sectors = ((MAX_CLIP_LENGTH * 2) / SST25_SECTOR_SIZE) + 1;
    int  tries = max_sectors + 2;
    
    while (tries-- != 0)
    {
        if (SDF_ferase(fidLayerBuffer) == 0)  break;  // done
    }
}


/*
 * Function sends data in sound clip buffer to console output stream,
 * formatted as C source code for an initialized array of 16-bit integers.
 * This is intended for applications in which PCM audio clips are stored in
 * MCU internal program flash memory.
 *
 * Command line format is: "synth -d [sound_clip_name]"
 * ... where <sound_clip_name> (if supplied) must comply with C identifier naming rules.
 *
 */
static void  DumpSoundBufferAs_C_sourceArray(int argCount, char *argVal[])
{
    char   outBuf[120];
    int16  sample, file_ID;
    int    i, column = 0;

    clipLength = (int) duration * 20;  // samples

    if (isOutputBufferEmpty) file_ID = fidLayerBuffer;
    else  file_ID = fidOutputBuffer;

    if (!SDF_fopen_RD(file_ID))
    {
        SDF_fclose();
        putstr("! Failed to open SDF buffer file.\n");
        return;
    }

    sprintf(outBuf, "// PCM sound clip data, length: %d samples, duration: %d ms\n",
            clipLength, (int) duration);
    putstr(outBuf);

    putstr("const int16  ");
    if (argCount == 3) putstr(argVal[2]);  // clip name
    else putstr("sound_name");
    putstr("[] = \n    { ");

    for (i = 0;  i < clipLength;  i++)
    {
        if ((column % 10) == 0) putstr("\n    ");
        sample = SDF_fread();  // LSB first
        sample = sample | ((int16) SDF_fread() << 8);  // MSB
        sprintf(outBuf, "%6d, ", sample);
        putstr(outBuf);
        if (++column >= 10)  column = 0;
    }

    SDF_fclose();
    putstr("\n    }; \n");
}


/*-----------------------------------------------------------------------------------------
 * Function writes contents of sound clip buffer to MDD media file (USB drive),
 * in raw "binary" format (array of 16-bit integers, little-endian byte order).
 *
 * Command line format is: "synth -s [clip_name]".
 *
 * The output file name is generated automatically; the format is "SF#####.RAW",
 * where ##### is a 5-digit random number.
 *
 * If <clip_name> is supplied, it will be written into a text file on MDD media
 * along with other data about the clip, e.g. duration, layer number, current layer
 * parameters, etc. The MDD file name will be the same as the clip RAW data file,
 * but with extension ".ID".
 */
static void  SaveSoundToMDDmediaRawFile(int argCount, char *argVal[])
{
    FSFILE *pFile;
    short  file_ID;    // SDF sound buffer file ID#
    uint32 fileNumber;
    char   filename[16];  // MDD '8.3' format
    char   soundClipName[24];
    char   outBuf[100];
    uint8  sectorBuf[SST25_SECTOR_SIZE];
    int    i, iData, bytesCopied = 0;
    int    fileSize, bytesBufd;

    clipLength = (int) duration * 20;  // samples

    if (!g_MDD_MediaMounted)
    {
        putstr("! USB drive not mounted.\n");
        return;
    }

    if (isOutputBufferEmpty) file_ID = fidLayerBuffer;
    else  file_ID = fidOutputBuffer;

    if (!SDF_fopen_RD(file_ID))
    {
        SDF_fclose();
        putstr("! Failed to open SDF buffer file.\n");
        return;
    }

    fileNumber = ((uint32) rand() + TMR1) % 100000;   // Must be 0 .. 99999
    sprintf(filename, "SF%05d.RAW", fileNumber);      // Random file-name

    pFile = FSfopen(filename, "w");
    if (pFile == NULL)
    {
        putstr("! Failed to open MDD output file: ");
        putstr(filename);
        putstr("\n");
        return;
    }

    fileSize = clipLength * 2;   // bytes

    while (bytesCopied < fileSize)
    {
        // Read a whole SDF file sector (4096 bytes) into RAM buffer
        for (bytesBufd = 0, i = 0;  i < SST25_SECTOR_SIZE;  i++)
        {
            iData = SDF_fread();
            if (iData < 0) break;    // unexpected end-of-file
            ////
            sectorBuf[i] = (uint8) iData;
            bytesBufd++;
            bytesCopied++;
            ////
            if (bytesCopied >= fileSize) break;  // end of clip data (last sector)
        }

        // Write (n = bytesBufd) bytes from buffer to MDD file
        if (FSfwrite(sectorBuf, bytesBufd, 1, pFile) != 1)
            break;
    }

    SDF_fclose();

    if (FSfclose(pFile) == ERROR)
        putstr("! Failed to write file: ");
    else
    {
        putDecimal(bytesCopied, 1);
        putstr(" bytes copied to file: ");
    }
    putstr(filename);  putstr("\n");

    if (argCount >= 3)   // Sound clip name supplied...  Create 'Sound ID' file
    {
        for (i = 0;  i < 19;  i++)  // get up to 19 printable chars
        {
            if (argVal[2][i] < 0x20)  break;  // Ctrl char found, e.g. NUL
            if (argVal[2][i] == ':') soundClipName[i] = '.';  // No colons allowed!
            else soundClipName[i] = argVal[2][i];
        }
        soundClipName[i] = 0;  // term char
        sprintf(outBuf, "Sound ID: %s, length: %d samples, duration: %d ms\n",
                soundClipName, clipLength, (int) duration);
        putstr(outBuf);

        sprintf(filename, "SF%05d.SID", fileNumber);
        pFile = FSfopen(filename, "w");

        if (pFile != NULL)
        {
            putstr("Writing sound info to file: ");  putstr(filename);  putstr("\n");

            FSfwrite(outBuf, 1, strlen(outBuf), pFile);  // Name, length, duration
            ListCurrentLayerParams(pFile);
            UpdateFSrealTimeVars();
            FSfclose(pFile);
        }
    }

    return;
}


/*-----------------------------------------------------------------------------------------
 * Function writes contents of sound clip buffer to MDD media file,
 * in Intel Hex (INHX32) format.
 *
 * Command line format is: "synth  -x  <sound_clip_name>"
 * ... where <sound_clip_name> may contain any printable chars except...
 *     SPACE, quotes, colon, [ ] / \ '&' * ? + = < >
 *
 * The Hex file name is generated automatically; the format is "PCM#####.HEX",
 * where ##### is a 5-digit random number.
 * The <sound_clip_name> is output in the 'title' record of the output file.
 *
 * INHX32 Hex record format: ":nnaaaattdd ... kk\r\n" (all digits are hex)
 *
 * where:  "nn"   is the record size (byte count),
 *         "aaaa" is the target address (LS 16 bits),
 *         "tt"   is the record type (00 = data, 01 = EOF, 04 = page, 09 = title),
 *         "kk"   is the record checksum (2's complement of sum of bytes),
 *         "dd"   is the first data byte (if rec type is "00").
 */
static void  SaveSoundToMDDmediaHexFile(int argCount, char *argVal[])
{
    char   outBuf[100];
    char   filename[16];        // MSDOS 8.3 format
    uint32 fileNumber;
    char   soundClipName[24];
    FSFILE *pFileHandle;
    int16  sample0, sample1;
    short  file_ID;
    int    i, isam = 0;
    int    recordLen, recordAddr;
    uint8  recordCksm;
    int    pageAddr, linearAddress = 0;  // for 32-bit linear address space

    clipLength = (int) duration * 20;  // number of samples

    if (!g_MDD_MediaMounted)
    {
        putstr("! USB drive not mounted.\n");
        return;
    }

    if (argCount >= 3)   // Sound clip name supplied...
    {
        for (i = 0;  i < 19;  i++)  // get up to 19 printable chars
        {
            if (argVal[2][i] < 0x20)  break;  // Ctrl char found, e.g. NUL
            if (argVal[2][i] == ':') soundClipName[i] = '.';  // No colons allowed!
            else soundClipName[i] = argVal[2][i];
        }
        soundClipName[i] = 0;  // term char
    }
    else   // No clip name given -- bail.
    {
        putstr("! Clip name must be supplied.\n");
        return;
    }

    if (isOutputBufferEmpty) file_ID = fidLayerBuffer;
    else  file_ID = fidOutputBuffer;

    if (!SDF_fopen_RD(file_ID))
    {
        SDF_fclose();
        putstr("! Failed to open SDF buffer file.\n");
        return;
    }

    fileNumber = ((uint32) rand() + TMR1) % 100000;   // Must be 0 .. 99999
    sprintf(filename, "PCM%05d.HEX", fileNumber);     // Random file-name

    UpdateFSrealTimeVars();
    pFileHandle = FSfopen(filename, "w");

    if (pFileHandle == NULL)
    {
        putstr("! Failed to open MDD media file: ");
        putstr(filename);
        putstr("\n");
        return;
    }

    sprintf(outBuf, "Clip length: %d samples, duration: %d ms, name: %s\n",
            clipLength, (int) duration, soundClipName);
    putstr(outBuf);

    putstr("Writing to MDD media file: ");
    putstr(filename);
    putstr(" ... \n");

    if (soundClipName[0] > 0x20)  // Valid sound clip name supplied...
    {
        // Output the clip name in a 'title' record (type 0x09)
        recordLen = strlen(soundClipName);  // byte count
        sprintf(outBuf, ":%02X%04X%02X[%s]\r\n", recordLen, 0x0000, 0x09, soundClipName);
//          putstr(outBuf);  // <<<<<<<<<<<<<<<<<<<< temp -- debug
        FSfwrite(outBuf, 1, strlen(outBuf), pFileHandle);
    }

    while (isam < clipLength)  // isam is the sample counter
    {
        recordAddr = linearAddress & 0xFFFF;  // start address for next DATA record
        if (recordAddr == 0x0000)  // Output a 'PAGE address' record...
        {
            pageAddr = linearAddress >> 16;  // A 'page' is 64kB (65536 bytes)
            recordCksm = 0x02 + HI_BYTE(pageAddr) + LO_BYTE(pageAddr) + 0x04;
            recordCksm = 0 - recordCksm;
            sprintf(outBuf, ":%02X%04X%02X%04X%02X\r\n",
                    0x02, 0x0000, 0x04, pageAddr, recordCksm);
//              putstr(outBuf);  // <<<<<<<<<<<<<<<<<<<< temp -- debug
            if (FSfwrite(outBuf, 1, strlen(outBuf), pFileHandle) != strlen(outBuf))
                break;  // write failed
        }

        if ((clipLength - isam) >= 16) recordLen = 32;  // bytes in data record, except..
        else  recordLen = (clipLength - isam) * 2;      // last data record
        recordCksm = recordLen + HI_BYTE(recordAddr) + LO_BYTE(recordAddr) + 0x00;

        // Write DATA record length, address & type info to buffer, thence out to hex file
        sprintf(outBuf, ":%02X%04X%02X", recordLen, recordAddr, 0x00);
//          putstr(outBuf);  // <<<<<<<<<<<<<<<<<<<< temp -- debug
        if (FSfwrite(outBuf, 1, strlen(outBuf), pFileHandle) != strlen(outBuf))
            break;  // write failed

        // Write up to 16 samples (32 data bytes) to the hex record
        for (i = 0;  i < (recordLen/2);  i++)
        {
            sample0 = SDF_fread();  // sample LSB
            sample1 = SDF_fread();  // sample MSB
            sprintf(&outBuf[i*4], "%02X%02X", sample0, sample1);
            recordCksm += sample0;
            recordCksm += sample1;
            isam++;
        }
        // Append checksum to buffer, then output the buffer to hex file
        recordCksm = 0 - recordCksm;
        i = strlen(outBuf);
        sprintf(&outBuf[i], "%02X\r\n", (int) recordCksm);
//          putstr(outBuf);  // <<<<<<<<<<<<<<<<<<<< temp -- debug
        if (FSfwrite(outBuf, 1, strlen(outBuf), pFileHandle) != strlen(outBuf))
            break;  // write failed

        linearAddress += 32;
    }

    if (isam == clipLength)
    {
        strcpy(outBuf, ":00000001FF\r\n");  // Output end-of-file record
//          putstr(outBuf);  // <<<<<<<<<<<<<<<<<<<< temp -- debug
        FSfwrite(outBuf, 1, strlen(outBuf), pFileHandle);
        putstr("Done.\n");
    }
    else putstr("\n! Error: Write failed.\n");

    SDF_fclose();
    UpdateFSrealTimeVars();
    FSfclose(pFileHandle);
}


// Set default values for user-settable synth parameters --------------
// All param's are stored as floats, but most are converted to integers at run-time.
//
static void SetDefaultParamValues(void)
{
    duration = 1000;
    patch = 1;
    osc1wave = SINE_WAVE;
    osc1freq = 1000;
    osc1level = 1.0;
    osc1offset = 0.0;
    osc2wave = SINE_WAVE;
    osc2freq = 500;
    osc2level = 0.0;
    osc2offset = 1.0;
    holdoff = 0;
    attack = 10;
    sustain = 900;
    release = 20;
    sweepbegf = 200;
    sweependf = 4000;
    sweeptime = 1000;
    sweephold = 1;
    prngseed = 0x0001;
    prngfbck = 0x1109;
}

