/*
 * Module:       wave_table_creator.c
 * ```````
 * Overview:     Command-line utility to generate and test wave-table data
 * `````````     for use with the "REMI" built-in sound synthesizer.
 *               The command syntax is:  wav  <option>  [args ...]
 *               Wave-table data is output in C source code format as an array definition
 *               to be included in source file "remi_synth_data.c".
 *
 * Originated:   M.J.Bauer, 2016   [www.mjbauer.biz/mjb_resources.htm]
 * ```````````
 * Reference:    "Wave-table Oscillator Notes" (text file in ../doc folder)
 * ``````````
 *=================================================================================================
 */

#include "system_def.h"
#include "wave_table_creator.h"

#include <math.h>

// Supported wave shapes for "partials"
#define  SINE_WAVE       1
#define  TRIANG_WAVE     2
#define  SQUARE_WAVE     3
#define  RAMP_WAVE       4
#define  NOISE_WAVE      5   // White noise

#define  CLIP_THRESHOLD   31900   // = 32 x 999

#define  PI  (3.141592654)

PRIVATE  void   ClearWaveTable(void);
PRIVATE  void   ListWaveParameters(char extras);
PRIVATE  void   OutputWaveformDescriptor(void);
PRIVATE  int    AddPartialToWaveTable(void);
PRIVATE  void   RemovePartialFromWaveTable(int order);
PRIVATE  void   ScaleTablePeakMagnitude(uint16 peak_pc);
PRIVATE  void   DumpWaveTable(uint8 format);
PRIVATE  int16  CalcWaveSample(uint8 wave_shape, int sample_point, int period);
PRIVATE  void   AmpldModulatePartial(int order, int depth); 
PRIVATE  void   RingModulateWave(short order);
PRIVATE  void   ApplyAntiAliasFilter();
PRIVATE  void   GenerateCustomWaveTable(int argc, char *argv[]);
PRIVATE  void   GenerateFreqWobulatorWave(int n1, int n2);
PRIVATE  void   GenerateChopperModWave(int n1, int n2);
PRIVATE  void   GeneratePseudoRandomWave(int srlen, int fbbits, int init);
PRIVATE  uint16 PseudoRandomSequence(uint8 srlen, uint16 seed, uint16 fbmask);


// Private data...
//
static  char   WaveTableName[42];    // Maximum 40 chars, plus NUL term-char
static  int    TableSize;            // Table size -- default 1260 samples, 2520 max. 
static  uint8  PartialAmpldHist[16]; // Histogram of partial amplitudes, [0] -> 1st order
static  int    WaveShape;            // Wave-shape (ID num) of partial
static  int    PartialAmpld;         // Amplitude of last partial added (0..100 %FS)
static  int    PartialOrder;         // Order of last partial added (maximum 12)
static  int    PartialPhase;         // Phase-shift of last partial added (samples)
static  uint16 AliasFilter_K;        // Filter time-constant (unit = sample intrvl)
static  bool   isTableEmpty;         // True when wave-table buffer is cleared
static  bool   isPrepDone = 0;       // True when 'wav' utility has been prepared for use
static  int    previousWaveTable;    // Wave-table in use prior to using 'wav' utility
static  int    previousPatch;        // Patch selected prior to using 'wav' utility


/*`````````````````````````````````````````````````````````````````````````````````````````````````
 *
 *  Function populates the wave-table buffer (array in data RAM) with sample data
 *  according to a specified waveform descriptor (arg1).
 *
 *  Entry arg:   waveDesc = pointer to a structure of type WaveformDesc_t
 */
void  GenerateWaveTable(WaveformDesc_t  *waveDesc)
{
    int   i;
    
    WaveShape = SINE_WAVE;
    PartialPhase = 0;
    TableSize = waveDesc->Size;
    ClearWaveTable();

    for (i = 0;  i < 12;  i++)
    {
        PartialOrder = i + 1;
        PartialAmpld = waveDesc->Partial[i];
        if (PartialAmpld != 0) AddPartialToWaveTable();
    }
}


/*^^^^
 * CLI command function:  Cmnd_wav
 *
 * The "wav" command provides facilities for the creation of wave-table data for use with
 * the oscillators in the "REMI" built-in sound synthesizer.
 * Sample values are signed 16-bit integers in the range +/-32000.
 *
 * A wave-table is populated by adding together one or more periodic waveforms called
 * "partials", of different relative frequency and phase. In the usual context, a "partial"
 * is a pure sine-wave, the frequency of which is not necessarily harmonically
 * related to the "fundamental" (dominant, usually lowest frequency, partial).
 *
 * In this application, the definition of "partial" may be broadened to include non-sinusoidal
 * waveforms such as square, triangle, ramp and "pitched" psuedo-random noise.
 * Be aware that non-sinusoidal partials can result in "aliasing" with the oscillator
 * sampling frequency, resulting in the generation of spurious unwanted frequencies.
 * Sinusoidal partials minimize the adverse effects of aliasing.
 *
 * Partial amplitudes are entered as percentages of full-scale, not as absolute values.
 * If the addition of a partial exceeds amplitude limits, i.e. if clipping occurs, a warning
 * will be issued by the "add" (-a) command.
 * It is recommended to plan the mix in advance so that the sum of the partial amplidudes
 * (percentages) is under 100%.  Note that the addition of partials does not generally
 * result in a waveform peak magnitude equal to the sum of the individual partial magnitudes.
 * This is particularly apparent when adding odd harmonics (i.e. 3rd, 5th, 7th order).
 *
 * The partial "order" is simply the harmonic number (1, 2, 3, 4, 5, ...) relative to the
 * "fundamental" (order = 1). One cycle of the fundamental fits the wave-table size exactly.
 * Choose a table size which results in an exact fit of partial period for all required
 * partials;  i.e. zero remainder when the table size is divided by the order, for all
 * partials with non-zero amplitude in the composite wave-form. Note that it is possible to
 * set zero amplitude for any partial, including the "fundamental" (order = 1).
 *
 * Recommended (default) table size is 1260, which fits most partial orders up to the 12th,
 * except the 8th and 11th orders, but if these partials have relatively low amplitude, the
 * resulting waveform distortion will not be preceptible.
 */
void  Cmnd_wav(int argCount, char * argVal[])
{
    char    option;
    char    textBuf[80];

    if (argCount == 1 || *argVal[1] == '?' )  // help
    {
        putstr( "Utility to create, test and save REMI synth wave-tables.  \n" );
        putstr( "Any cmd option except '-x' will activate User wave-table. \n" );
        putstr( "Enter 'wav -x' cmd when done to restore previous patch. \n\n" );
        putstr( "Usage:  wav  <option>  [arg1] [arg2] ... \n" );
        putstr( "<option>-----------------------------------------\n" );
        putstr( "  -n : New wave-table  [arg1 = table size]       \n" );
        putstr( "  -f : Set Frequency Divider = <arg1>            \n" );
        putstr( "  -l : List wave-table and partial param's       \n" );
        putstr( "  -a : Add partial to wave-table buffer --       \n" );
        putstr( "           <arg1> = partial order (1..12)        \n" );
        putstr( "           <arg2> = partial ampld (1..99 %FS)    \n" );
//      putstr( "           [arg3] = partial phase (0..360 deg)   \n" );  // todo ?
        putstr( "  -c : Clear partial, <arg1> = order (1..12)     \n" );
        putstr( "  -z : Scale table peak magnitude [arg = %FS]    \n" );
        putstr( "  -d : Output waveform Descriptor --             \n" );
        putstr( "           <arg1> = name (no spaces), optional   \n" );
        putstr( "  -o : Output wave-table as C array def'n        \n" );
        putstr( "  -s : Save user wave-table def'n in EEPROM      \n" );
        putstr( "  -x : Exit -- restore last patch selected       \n" );
        putstr( "-------------------------------------------------\n" );
        return;
    }

    if (strmatch(argVal[1], "++"))  // show "undocumented" wav commands
    {
        putstr( "\n" );
        putstr( "  -g : Generate custom waveform  (see usage)     \n" );
        putstr( "  -w : Set Waveform of 'partial' --              \n" );
        putstr( "           1 = sine, 2 = triangle, 3 = square,   \n" );
        putstr( "           4 = ramp, 5 = random noise            \n" );
        putstr( "  -k : Apply anti-alias filter (K = 5..500)      \n" );
        putstr( "  -m : Ampld-Modulate partial with buffer wave   \n" );
        putstr( "           <arg1> = partial order (1..12)        \n" );
        putstr( "           <arg2> = modulation depth (1..200%)   \n" );
        putstr( "  -r : Ring-modulate buffer wave with partial    \n" );
        putstr( "           <arg1> = partial order (1..12)        \n" );
        putstr( "  -t : Play test tone using REMI synth...        \n" );
        putstr( "           <arg1> = MIDI note number (23..108)   \n" );
        putstr( "           <arg2> = duration (seconds)           \n" );
        putstr( "-------------------------------------------------\n" );
        return;
    }

    if (argVal[1][0] == '-') option = tolower(argVal[1][1]);
    else option = '$';  // undefined

    // One-time initialization on first use of 'wav' command, except option '-x'
    if (!isPrepDone && option != 'x' && option != '$')
    {
        previousPatch = GetActivePatchID();  // Save active patch before corrupting it
        SelectUserWaveTableOsc1();           // Use RAM buffer and disable OSC2
        TableSize = 1260;                    // Set defaults for user wave-table
        WaveTableSizeSet(TableSize);
        Osc1FreqDividerSet(1.0);
        WaveShape = SINE_WAVE;
        PartialPhase = 0;
        AliasFilter_K = 50;
        ClearWaveTable();
        strcpy(WaveTableName, "Untitled_wave");
        srand(TMR2 | 1);                       // Seed rand() with an odd number
        isPrepDone = 1;
    }

    switch (option)
    {
    case 'n':           // New wave-table
    {
        int     arg1 = atoi(argVal[2]);

        if (argCount == 2)  // size not specified - use default
        {
            putstr("Table size not specified - using default: 1260 \n");
            TableSize = 1260;
        }
        else if (arg1 >= 200 && arg1 <= WAVE_TABLE_MAXIMUM_SIZE)
        {
            TableSize = arg1;
        }
        else  // size out of range
        {
            sprintf(textBuf, "! Table size must be 200..%d samples. \n", WAVE_TABLE_MAXIMUM_SIZE);
            putstr(textBuf);
            break;
        }

        WaveTableSizeSet(TableSize);
        Osc1FreqDividerSet(1.0);
        strcpy(WaveTableName, "Untitled_waveform");
        WaveShape = SINE_WAVE;
        PartialPhase = 0;
        ClearWaveTable();
        putstr("Output buffer is empty... Ready to add partial.\n");
        break;
    }
    case 'f':           // Set Osc. Freq. Divider
    {
        float   arg = atof(argVal[2]);

        if (argCount == 3 && arg >= 0.001)  Osc1FreqDividerSet(arg);

        sprintf(textBuf, "Freq. Divider: %8.4f \n", Osc1FreqDividerGet());
        putstr(textBuf);
        break;
    }
    case 'l':           // List table & partial stat's
    {
        bool  extras = (argCount > 1);

        ListWaveParameters(extras);
        break;
    }
    case 'a':           // Add partial to wave-table buffer
    {
        int   order = atoi(argVal[2]);
        int   ampld = atoi(argVal[3]);
        int   peak_ampld;

        if (order < 1 || order > 12)
        {
            putstr("! Invalid order - Partial not generated. \n");
            break;
        }

        if (ampld <= 0 || ampld > 99)
        {
            putstr("! Ampld must be 1..99 (%) - Partial not generated.\n");
            break;
        }

//      if ((TableSize % order) != 0)
//          putstr("! Table size is not an exact multiple of partial period.\n");

        PartialOrder = order;
        PartialAmpld = ampld;
        peak_ampld = AddPartialToWaveTable();

        putstr("* Partial order ");
        putDecimal(order, 1);
        putstr(" added to wave-table.\n");

        if (peak_ampld == ERROR)
            putstr("! Clipping occurred in composite waveform.\n");
        else
        {
            putstr("  Waveform peak amplitude: ");
            putDecimal(peak_ampld, 1);
            putstr(" (");
            putDecimal((100 * peak_ampld) / 32000, 1);
            putstr("% FS) \n");
        }
        break;
    }
    case 'c':           // Clear partial from wave-table
    {
        int   order = atoi(argVal[2]);

        if (order < 1 || order > 12)
        {
            putstr("! Invalid order - nothing changed. \n");
            break;
        }

//      PartialAmpld = PartialAmpldHist[order-1];
//      if (PartialAmpld != 0)  SubtractPartialFromWaveTable();
        
        RemovePartialFromWaveTable(order);
        putstr("* Partial removed from waveform.\n");
        break;
    }
    case 'z':           // Scale samples in wave buffer
    {
        uint16   peak_pc = atoi(argVal[2]);

        if (argCount < 3 || peak_pc > 100)  peak_pc = 95;

        if (!isTableEmpty)
        {
            ScaleTablePeakMagnitude(peak_pc);
            ListWaveParameters(0);
        }
        else  putstr("! Buffer is empty. \n");
        break;
    }
    case 'd':           // Output waveform descriptor
    {
        if (argCount >= 3 && strlen(argVal[2]) > 0)  // name supplied
        {
            strncpy(WaveTableName, argVal[2], 40);
            WaveTableName[40] = 0;  // NUL term-char
        }

        OutputWaveformDescriptor();

        if (isTableEmpty)  putstr("! Buffer is empty. \n");
        break;
    }
    case 'o':           // Output wave-table as C array def'n
    {
        uint8  format = 0;

        if (argCount >= 3 && strlen(argVal[2]) > 0)  // name supplied
        {
            strncpy(WaveTableName, argVal[2], 40);
            WaveTableName[40] = 0;  // NUL term-char
        }

        if (WaveShape == SINE_WAVE)  format = 1;

        if (!isTableEmpty)  DumpWaveTable(format);
        else  putstr("! Buffer is empty. \n");
        break;
    }
    case 's':           // Save User WaveTable param's in EEPROM
    {
        int   i;

        if (isTableEmpty)
        {
            putstr("! Buffer is empty. \n");
            break;
        }

        g_Config.UserWaveform.Size = TableSize;
        g_Config.UserWaveform.FreqDiv = Osc1FreqDividerGet();

        for (i = 0;  i < 12;  i++)
        {
            g_Config.UserWaveform.Partial[i] = PartialAmpldHist[i];
        }

        if (StoreConfigData())
            putstr("* User Wave-table param's saved OK.\n");
        else  putstr("! Error: EEPROM write failed.\n");
        break;
    }
    case 'x':           // Exit 'wav' utility, restore previous patch
    {
        RemiSynthPatchSetup(previousPatch);
        isPrepDone = FALSE;
        break;
    }
    //
    // ------- Following are "undocumented" (hidden) command options ----------
    //
    case 'g':           // Generate custom wave-table
    {
        GenerateCustomWaveTable(argCount, argVal);
        break;
    }
    case 'w':           // Set wave shape of "partial" to be processed
    {
        int   arg = atoi(argVal[2]);

        if (arg >= 1 && arg < 6)  WaveShape = arg;
        else putstr("! Undefined wave-shape ID. \n");
        break;
    }
    case 'k':           // Apply anti-alias filter, time-constant K = arg1
    {
        int  k = 50;

        if (argCount >= 3) k = atoi(argVal[2]);
        else  putstr("! Missing arg. - using default (K = 50).\n");
        
        if (k <= 0 || k > 500)  
            putstr("! Arg out of bounds. - Filter not applied.\n");
        else
        {
            AliasFilter_K = k;
            ApplyAntiAliasFilter();
            putstr("* Filter applied OK. Normalization (-z) recommended.\n");
        }
        break;
    }
    case 'h':           // Set phase-shift of partial (in samples)
    {
        int   phase = atoi(argVal[2]);

        if (phase < 0 || phase >= TableSize)
        {
            putstr("! Phase must be >= zero and < table size. \n");
            break;
        }

        PartialPhase = phase;
        break;
    }
    case 'm':           // Ampld-Modulate a partial with buffered waveform
    {
        int   partialOrder;  // partial order (carrier)
        int   depth;         // mod'n index (%)

        if (argCount < 4)
        {
            putstr("! Missing arg(s). \n");
            break;
        }
        
        partialOrder = atoi(argVal[2]);
        depth = atoi(argVal[3]);
        
        if (partialOrder < 1 || partialOrder > 12 || depth < 1 || depth > 200)
        {
            putstr("! Arg value out of bounds. \n");
            break;
        }

        AmpldModulatePartial(partialOrder, depth); 

        break;
    }
    case 'r':           // Ring-modulate the buffered waveform with a partial
    {
        int   partialOrder = atoi(argVal[2]);

        if (argCount < 3 || partialOrder < 1 || partialOrder > 12)
            putstr("! Missing or invalid arg: partial order (1..12) \n");
        else  RingModulateWave(partialOrder);
        break;
    }
    case 't':           // Test sound of buffered waveform using REMI synth
    {
        static  int    noteNumber = 69;        // default = A4, 440Hz
        static  int    noteDuration = 2000;    // default = 2000ms (2 sec)
        static  uint32 testEndTime;

        int  arg1 = atoi(argVal[2]);           // Osc. frequency (Hz)
        int  arg2 = atoi(argVal[3]);           // Tone duration (sec)

        if (isTableEmpty)
        {
            putstr("! Buffer is empty. \n");
            break;
        }

        if (argCount >= 3 && arg1 >= 20 && arg1 <= 120) noteNumber = arg1;
        if (argCount >= 4 && arg2 > 0) noteDuration = arg2 * 1000;  // msec

        testEndTime = millisecTimer() + noteDuration;
        RemiSynthNoteOn(noteNumber, 64);
        RemiSynthExpression(64);

        while (millisecTimer() < testEndTime)
        {
            BackgroundTaskExec();
        }

        RemiSynthNoteOff();
        break;
    }

    default:
        putstr("! Undefined cmd option.\n");
        break;

    } // end switch
}


PRIVATE  void  ClearWaveTable(void) 
{
    int   i;

    for (i = 0;  i < TableSize;  i++)
    {
        WaveTableBuffer[i] = 0;
    }

    for (i = 0;  i < 16;  i++)
    {
        PartialAmpldHist[i] = 0;
    }

    isTableEmpty = TRUE;
}


/*
 * Function outputs wave-table parameters in human-readable form.
 * Called by "wav -l" command option.
 */
PRIVATE  void   ListWaveParameters(char extras)
{
    int   order, i;
    int   absValue, peakAmpld;
    char  textBuf[80];
    bool  isAdditiveWave = 0;  

    putstr("Table name: ");  putstr(WaveTableName);  putstr("\n");
    putstr("Table size: ");  putDecimal(TableSize, 1);  putstr(" samples \n");
    sprintf(textBuf, "Osc Freq Divider: %8.4f \n", Osc1FreqDividerGet());
    putstr(textBuf);
    putstr("Anti-alias filter: K = ");
    putDecimal(AliasFilter_K, 1);  putstr("\n");

    if (isTableEmpty)  
    {
        putstr("! Buffer is empty. \n");
        return;
    }
    
    for (order = 1;  order <= 12;  order++)
    {
        if (PartialAmpldHist[order-1] != 0) isAdditiveWave = 1;
    }
    
    if (isAdditiveWave)  // waveform is sum of partials
    {
        putstr("Partial waveform: ");
        if (WaveShape == SINE_WAVE)  putstr("Sine wave \n");  else
        if (WaveShape == TRIANG_WAVE)  putstr("Triangle \n");  else
        if (WaveShape == SQUARE_WAVE)  putstr("Square \n");  else
        if (WaveShape == RAMP_WAVE)  putstr("Sawtooth \n");  else
        if (WaveShape == NOISE_WAVE)  putstr("Noise \n");
        
        putstr("Partial Distribution: \n");
        putstr("Order # |  1 |  2 |  3 |  4 |  5 |  6 |  7 |  8 |  9 | 10 | 11 | 12 | \n");
        putstr("Ampld % | ");

        for (order = 1;  order <= 12;  order++)
        {
            if (PartialAmpldHist[order-1] != 0)
                putDecimal(PartialAmpldHist[order-1], 2);
            else  putstr("  ");
            putstr(" | ");
        }
        putstr("\n");
    }
    
    // This loop finds the peak magnitude of samples in the wave-table
    for (peakAmpld = 0, i = 0;  i < TableSize;  i++)
    {
        absValue = abs(WaveTableBuffer[i]);
        if (absValue > peakAmpld) peakAmpld = absValue;
    }
    
    sprintf(textBuf, "Peak ampld: %d (%d %%FS) \n", peakAmpld, (peakAmpld * 100) / 32000);
    putstr(textBuf);
}


/*
 * Function outputs waveform descriptor to console in C source format, ready to copy
 * and paste into the wave-table definitions array in file "remi_synth_data.c".
 * Called by "wav -d" command option.
 */
PRIVATE  void   OutputWaveformDescriptor(void)
{
    int   order;
    char  textBuf[80];

    putstr("\n");
    putstr("    {   // ID# xx: ");  putstr(WaveTableName);  putstr("\n");
    putstr("        ");  putDecimal(TableSize, 1);  putstr(",\n");
    putstr("        ");
    sprintf(textBuf, "%5.3f,\n", Osc1FreqDividerGet());
    putstr(textBuf);
    putstr("        { ");

    for (order = 1;  order <= 12;  order++)
    {
        putDecimal(PartialAmpldHist[order-1], 1);
        if (order < 12) putstr(", ");
    }
    putstr(" } \n");
    putstr("    },\n\n");
}


/*
 * Function adds the specified partial to the wave-table output buffer.
 * Called by "wav -a" command option.
 *
 * It is assumed that the required partial wave parameters have been set up prior, i.e.
 *     PartialOrder, PartialAmpld, PartialPhase.
 *
 * Return val:  (int)  positive peak sample value of updated wave-table, or
 *                     ERROR (-1) if clipping occurred.
 *                     Clipping threshold is +/-31900 (= 32 x 999 approx.)
 */
PRIVATE  int  AddPartialToWaveTable(void)
{
    int   sample_val, partial_val;
    int   new_sample_val, peak_val = 0;
    int   i = 0;
    int   order = PartialOrder;
    int   period = TableSize/PartialOrder;
    int   j = PartialPhase;       // Partial wave is phase-shifted by j sample points

    for (i = 0;  i < TableSize;  i++, j++)
    {
        if (j >= TableSize) j = 0;  // wrap

        sample_val = CalcWaveSample(WaveShape, i, period);
        partial_val = (sample_val * PartialAmpld) / 100;
        new_sample_val = WaveTableBuffer[j] + partial_val;

        if (new_sample_val < -CLIP_THRESHOLD)
        {
            WaveTableBuffer[j] = -CLIP_THRESHOLD;
            peak_val = ERROR;
        }
        else if (new_sample_val > CLIP_THRESHOLD)
        {
            WaveTableBuffer[j] = CLIP_THRESHOLD;
            peak_val = ERROR;
        }
        else  WaveTableBuffer[j] = new_sample_val;

        if (peak_val != ERROR && new_sample_val > peak_val)
            peak_val = new_sample_val;
    }

    // Update the Partial Histogram array...
    if (order <= 12 && (PartialAmpldHist[order-1] + PartialAmpld) < 200)
        PartialAmpldHist[order-1] += PartialAmpld;
    else  PartialAmpldHist[order-1] = 200;  // Limit

    isTableEmpty = FALSE;
    return  peak_val;
}


/*
 * Function effectively removes the specified partial from the wave-table buffer.
 * Called by "wav -c <n>" (Clear) command option.
 *
 * The buffered waveform is regenerated using the partial distribution stored in the
 * array PartialAmpldHist[], except that the unwanted partial has ampld set to zero.
 * 
 * All partials are assumed to be sinusoidal.
 */
PRIVATE  void  RemovePartialFromWaveTable(int order)
{
    int  i;
    
    ClearWaveTable();
    PartialPhase = 0;
    WaveShape = SINE_WAVE;
    PartialAmpldHist[order-1] = 0;  // unwanted partial

    for (i = 0;  i < 12;  i++)
    {
        PartialOrder = i + 1;
        PartialAmpld = PartialAmpldHist[i];
        if (PartialAmpld != 0) AddPartialToWaveTable();
    }
}


/*
 * Function calculates and returns a sample value in a periodic waveform at a given point
 * in the cycle.
 *
 * Entry arg's: wave_shape   = waveform ID code (sine = 1, triangle = 2, etc)
 *              sample_point = wave angle relative to period as a number of sample points
 *                               e.g. period/4 represents an angle of 90 deg.
 *                               sample_point may be higher than period.
 *              period       = value of sample_point at wave angle = 360 deg.
 *
 * Return val:  (int16)  sample_value, signed 16-bit integer.
 *                       Full-scale (peak value) is +/-32000 for all wave-shapes.
 */
PRIVATE  int16  CalcWaveSample(uint8 wave_shape, int sample_point, int period)
{
    static  int32   sample_value = 0;
    int16   noise_value;
    int     index;

    sample_point = sample_point % period;  // in case sample_point > period

    if (wave_shape == SQUARE_WAVE)
    {
        sample_value = (sample_point < (period/2)) ? 32000 : -32000;
    }
    else if (wave_shape == TRIANG_WAVE)
    {
        if (sample_point < (period/4))  // up to 90 deg.
            sample_value = (32000 * sample_point) / (period/4);
        else if (sample_point < (period/2))  // more than 90 deg, up to 180 deg.
            sample_value = 32000 - (32000 * (sample_point - period/4)) / (period/4);
        else if (sample_point < (3 * period) / 4)  // more than 180 deg, up to 270 deg
            sample_value = 0 - (32000 * (sample_point - period/2)) / (period/4);
        else  // more than 270 deg, up to 360 deg.
            sample_value = (32000 * (sample_point - (3 * period) / 4)) / (period/4) - 32000;
    }
    else if (wave_shape == RAMP_WAVE)
    {
        sample_value = (2 * 32000 * sample_point) / period - 32000;
    }
    else if (wave_shape == NOISE_WAVE)
    {
        do  { noise_value = (int16)(rand() & 0xFFFF); }
        until (noise_value > -32000 && noise_value < 32000);  // limits
    }
    else  // if (wave_shape == SINE_WAVE)
    {
        if (period == SINE_WAVE_TABLE_SIZE)  // fastest execution
            sample_value = g_sinewave[sample_point];
        else
        {
            index = (SINE_WAVE_TABLE_SIZE * sample_point) / period;
            sample_value = g_sinewave[index];
        }
    }

    return  (int16) sample_value;
}


/*
 * Function scales all sample values in the wave-table buffer, so that the peak
 * magnitude is a given percentage of full-scale (FS = 32000). If no peak value is
 * supplied (arg == 0), then the peak magnitude will be 30000 (approx. 95% FS).
 * The Partial Histogram array is scaled by the same factor.
 *
 * Called by "wav -z" (normalize) command option.
 */
PRIVATE  void  ScaleTablePeakMagnitude(uint16 peakPercent)
{
    int    i;
    int    abs_value;
    int    peakTarget;
    int    peakAmpld = 0;      // peak magnitude of samples
    float  scaleFactor;

    // This loop finds the peak magnitude of samples in the wave-table
    for (i = 0;  i < TableSize;  i++)
    {
        abs_value = abs(WaveTableBuffer[i]);
        if (abs_value > peakAmpld) peakAmpld = abs_value;
    }

    if (peakAmpld != 0)
    {
        peakTarget = 32000 * peakPercent / 100;
        scaleFactor = (float) peakTarget / peakAmpld;

        for (i = 0;  i < TableSize;  i++)
        {
            WaveTableBuffer[i] = (int16) (scaleFactor * WaveTableBuffer[i]);
        }

        for (i = 0;  i < 16;  i++)    // Adjust partial histogram
        {
            PartialAmpldHist[i] = (float) PartialAmpldHist[i] * scaleFactor;
            if (PartialAmpldHist[i] > 99) PartialAmpldHist[i] = 99;
        }
    }
}


/*
 * Function:    ApplyAntiAliasFilter()
 *
 * The buffered waveform is low-pass filtered to remove high-order partials which might
 * otherwise cause aliasing with the sample frequency.  The filter works by averaging a
 * finite number of consecutive samples (= AliasFilter_K).
 */
PRIVATE  void   ApplyAntiAliasFilter()
{
    int16   saveBuf[512];
    int32   sum;
    int     i, j, k;

    if (isTableEmpty)
    {
        putstr("! Buffer is empty. \n");
        return;
    }

    for (k = 0;  k < AliasFilter_K;  k++)  // Save the first K samples
    {
        saveBuf[k] = WaveTableBuffer[k];
    }

    for (i = 0;  i < TableSize;  i++)
    {
        // Find the sum of K consecutive samples at WaveTableBuffer[i]
        for (sum = 0, k = 0;  k < AliasFilter_K;  k++)
        {
            j = i + k;
            if (j < TableSize)  sum += WaveTableBuffer[j];
            else  sum += saveBuf[k];  // use preserved value
        }
        WaveTableBuffer[i] = sum / AliasFilter_K;  // filtered sample = mean
    }
}


/*
 * Function outputs the wave-table as an array of sample values (decimal integers),
 * in C source syntax, i.e. initialized array of constants.
 * Called by "wav -c" command option. Intended for development purposes only.
 *
 * Entry arg. format = 0 for brief, 1 to show partial distribution
 */
PRIVATE  void  DumpWaveTable(uint8 format)
{
    char   outBuf[80];
    int16  sample;
    int    i, order, column = 0;

    // Print wave-table info as C comment block
    putstr("\n\n/* \n");
    putstr(" * Wave-table definition ... \n");
    sprintf(outBuf, " * Waveform name: %s, size: %d samples \n",
            WaveTableName, TableSize);
    putstr(outBuf);
    sprintf(outBuf, " * Osc. Freq. Divider: %8.4f \n", Osc1FreqDividerGet());
    putstr(outBuf);

    if (format == 1)  // Assume all partials are sinusoidal
    {
        putstr(" * Waveform Partial Distribution: \n");
        putstr(" * Order # |  1 |  2 |  3 |  4 |  5 |  6 |  7 |  8 |  9 | 10 | 11 | 12 | \n");
        putstr(" * Ampld % | ");

        for (order = 1;  order <= 12;  order++)
        {
            if (PartialAmpldHist[order-1] != 0)
                putDecimal(PartialAmpldHist[order-1], 2);
            else  putstr("  ");
            putstr(" | ");
        }
        putstr("\n");
    }
    putstr(" */ \n");

    sprintf(outBuf, "const  int16  %s[] = \n{ ", WaveTableName);
    putstr(outBuf);

    for (i = 0;  i < TableSize;  i++)
    {
        if ((column % 10) == 0) putstr("\n    ");
        sample = WaveTableBuffer[i];
        sprintf(outBuf, "%6d", sample);
        putstr(outBuf);
        if (i < (TableSize - 1)) putstr(", ");
        if (++column >= 10)  column = 0;
    }

    putstr("\n}; \n\n");
}


/*
 * This function amplitude-modulates a specified partial ("carrier" frequency) 
 * with the buffered waveform (scaled and DC offset to 0.5 FS).
 * 
 * The partial order (1..12) and modulation depth (%) are specified by command arg's.
 * The modulated partial (carrier) waveform is always sinusoidal.
 */
PRIVATE  void   AmpldModulatePartial(int partialOrder, int depth)
{
    int32   partial_val;  // sample in modulated wave (sinusoidal "carrier")
    int32   mod_wave;     // sample in modulating wave (buffered waveform)
    int32   sample_val;   // sample in output waveform
    int     i;
    int     period = TableSize / partialOrder;

    if (isTableEmpty)
    {
        putstr("! Buffer is empty. \n");
        return;
    }

    for (i = 0;  i < TableSize;  i++)
    {
        partial_val = CalcWaveSample(SINE_WAVE, i, period);
        mod_wave = ((int) WaveTableBuffer[i] * depth) / 200 + 16000;
        sample_val = (partial_val * mod_wave) >> 16;
        
        if (sample_val > 32000) sample_val = 32000;    // Apply limiter
        if (sample_val < -32000) sample_val = -32000;  
        WaveTableBuffer[i] = (int16) sample_val;
    }
}


/*
 * This function ring-modulates the pre-buffered waveform by a specified partial.
 * The partial order is specified by the command argument.
 * Other partial parameters (waveshape, phase, etc) are as currently set.
 *
 * If the buffered waveform is noise, it should have been filtered by the anti-alias filter
 * using a time-constant (K) chosen so the corner frequency (Fc) relative to the fundamental
 * (Fo) of the output waveform is optimum for the modulating partial frequency.
 *
 * Example:
 * Let's populate the wave-table with 6 cycles of the modulating frequency (Fm).
 * Let's set the filter corner freq to coincide with Fm, i.e. Fc = Fm = 6 * Fo.
 * The filter time-constant is given by Tc = 1 / (2 * pi * Fc).
 * The period of Fm is 2520 / 6 = 420 samples. Hence Tc = 420 / 6 = 70 samples.
 * Therefore, K = 70 samples.
 */
PRIVATE  void   RingModulateWave(short partialOrder)
{
    int32   sample_val, partial_val;
    int     i;
    int     order = partialOrder;   // order of modulating partial
    int     period = TableSize / order;

    if (isTableEmpty)
    {
        putstr("! Buffer is empty. \n");
        return;
    }

    for (i = 0;  i < TableSize;  i++)
    {
        partial_val = CalcWaveSample(SINE_WAVE, i, period);
        sample_val = (WaveTableBuffer[i] * partial_val) >> 15;
        WaveTableBuffer[i] = (int16) sample_val;
    }
}


/*
 * Function to generate odd-ball waveforms.
 *
 * Entry arg's specify the waveform parameters.
 */
PRIVATE  void   GenerateCustomWaveTable(int argCount, char * argVal[])
{
    int  n1, n2, n3;

    if (argCount < 4 )  // help needed
    {
        putstr( "Usage (1):  wav  -gf  <n1> <n2> \n" );
        putstr( "  - Generates an FM waveform with frequency wobulation \n" );
        putstr( "  - n1 & n2 are partial orders of Fmin and Fmax (1..20) \n" );
        putstr( "Usage (2):  wav  -gc  <n1> <n2> \n" );
        putstr( "  - Generates a chopper-modulated waveform with 2 freq's \n" );
        putstr( "  - n1 & n2 are partial orders of F1 and F2 (1..20) \n" );
        putstr( "Usage (3):  wav  -gr  <nb> <fbm> [seed] \n" );
        putstr( "  - Generates a binary pseudo-random sequence waveform \n" );
        putstr( "  - nb = SR size, bits (5..15) \n" );
        putstr( "  - fbm = feedback bitmask, hex (1..3FFF) \n" );
        putstr( "  - seed = initial SR data, hex (1..3FFF), dflt = 1 \n" );
        return;
    }

    n1 = atoi(argVal[2]);
    n2 = atoi(argVal[3]);

    if (strmatch(argVal[1], "-gf"))
    {
        if (n1 < 1 || n2 <= n1)
        {
            putstr("! Must have N1 >= 1, N2 > N1. \n");
            return;
        }
        GenerateFreqWobulatorWave(n1, n2);
        isTableEmpty = FALSE;
    }
    else if (strmatch(argVal[1], "-gc"))
    {
        if (n1 < 1 || n2 <= n1)
        {
            putstr("! Must have N1 >= 1, N2 > N1. \n");
            return;
        }
        GenerateChopperModWave(n1, n2);
        isTableEmpty = FALSE;
        
        if ((n1 % 2) != 0 && (n2 % 2) == 0)
			putstr("! Attn: N1 is odd, N2 is even -> asymmetrical waveform.\n");
        if ((n1 % 2) == 0 && (n2 % 2) != 0)
			putstr("! Attn: N1 is even, N2 is odd -> asymmetrical waveform.\n");
    }
    else if (strmatch(argVal[1], "-gr"))
    {
        if (n1 < 5 || n1 > 15)
        {
            putstr("! Arg(s) out of bounds. \n");
            return;
        }
        n2 = hexatoi(argVal[3]);
        if (argCount == 5) n3 = hexatoi(argVal[4]);  else n3 = 1;

        GeneratePseudoRandomWave(n1, n2, n3);
        isTableEmpty = FALSE;
        
        putstr( "* Anti-alias filtering (-k) recommended. \n" );
    }
    else  putstr("! Invalid cmd option. \n");
}


PRIVATE  void  GenerateFreqWobulatorWave(int n1, int n2)
{
    int    i;
    float  fnorm;          // Instantaneous output freq, normalized to Fo
    float  fmax, fmin;     // Min & max values of fnorm
    float  alpha;          // Wave angle (0..2*PI) at f = Fo
    float  theta;          // Wave angle at sample point (max = 2*PI*fnorm)
    float  delta;          // Change in theta from successive sample points
    int16  sample;

    fmin = (float) n1;
    fmax = (float) n2;

    for (i = 0;  i < TableSize;  i++)
    {
        alpha = 2*PI * ((float) i / TableSize);
        fnorm = (fmax + fmin) / 2 + ((fmax - fmin) / 2) * cosf(alpha);
		sample = (int16) (30000 * sinf(theta));
		delta = (2*PI * fnorm) / TableSize;
		theta = theta + delta;
        WaveTableBuffer[i] = sample;
    }
}


PRIVATE  void  GenerateChopperModWave(int n1, int n2)
{
    int    i, j;
    float  alpha;          // Wave angle (0..2*PI) at f = Fo
	int16  sample;

    for (i = 0;  i < (TableSize / 2);  i++)   // 1st half of table
    {
		alpha = 2*PI * n1 * ((float) i / TableSize);
		sample = (int16) (30000 * sinf(alpha));
        WaveTableBuffer[i] = sample;
    }

	for (i = 0;  i < (TableSize / 2);  i++)   // 2nd half of table
    {
		alpha = 2*PI * n2 * ((float) i / TableSize);
		if ((n1 % 2) != 0)  alpha += PI;   // N1 is odd -- invert N2 wave
		sample = (int16) (30000 * sinf(alpha));
		j = (TableSize / 2) + i;
        WaveTableBuffer[j] = sample;
    }
}


PRIVATE  void  GeneratePseudoRandomWave(int srlen, int fbbits, int init)
{
    int     i;
	uint16  so;     // SR serial out (bit_0)
	int16   sample;

	if (srlen < 5) srlen = 5;
	if (srlen > 15) srlen = 15;
    if (init <= 0) init = 1;

	PseudoRandomSequence(srlen, init, fbbits);  // Initialize the sequence

    for (i = 0;  i < TableSize;  i++)
    {
		so = PseudoRandomSequence(srlen, 0, fbbits) & 1;
		if (so == 0)  sample = -16000;  else sample = 16000;
        WaveTableBuffer[i] = sample;
    }
}


/*
 * Pseudo-Random Sequence Generator using Shift-Register with Feedback.
 * Register length is variable, 5 to 15 bits, set by arg1 (srlen).
 * The register is "clocked" (shifted 1 bit posn) on each call.
 * To generate a binary sequence, use only bit_0 of the return value.
 *
 * Entry arg's: (uint8)  srlen  = shift-register length (5..15 bits)
 *
 *              (uint16) seed   = initial shiftreg data if non-zero (0..0x7FFF)
 *                                ... shiftreg unchanged if seed is zero.
 * 
 *              (uint16) fbmask = feedback bit mask (0..0x3FFF)
 *
 * Return val:  (uint16) Next number in sequence, unsigned integer
 */
PRIVATE  uint16  PseudoRandomSequence(uint8 srlen, uint16 seed, uint16 fbmask)
{
    static  uint32  shiftreg;  // shift register (5..15 LS bits)
    int     i;
    uint32  exorbits;  // bits to be XOR'd together to feed shiftreg input (si)
    uint32  si = 1;    // shiftreg serial input bit
	uint32  carrybit = (1 << srlen);  // bit position of si in shiftreg

    if (seed != 0)  // one-time initialization
    {
        shiftreg = (uint32) seed;
    }
    else  // normal running
    {
        exorbits = shiftreg & fbmask;

        for (i = 0;  i < srlen;  i++)   // calculate si = XOR of feedback bits
        {
            si = si ^ (exorbits & 1);
            exorbits = exorbits >> 1;
        }

        shiftreg &= ~carrybit;   // Clear carry bit
        if (si == 1) shiftreg |= carrybit;  // Set carry bit
        shiftreg = shiftreg >> 1;  // Shift right 1 bit (unsigned)
    }

    return  (uint16) shiftreg;
}


/*=============================  LICENSE AGREEMENT  ===================================*\
 *
 *  THIS SOURCE CODE MAY BE USED FREELY FOR PERSONAL NON-COMMERCIAL APPLICATIONS.
 *  HOWEVER, THE CODE OR PART THEREOF MAY NOT BE REDISTRIBUTED OR SOLD IN ANY FORM.
 *  USE OF THIS SOURCE CODE OR OBJECT CODE GENERATED FROM IT FOR COMMERCIAL PURPOSES
 *  WITHOUT THE EXPRESS WRITTEN PERMISSION OF THE AUTHOR IS PROHIBITED.
 *
\*=====================================================================================*/

